import os
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from prefect import Flow, task

from viadot.task_utils import add_ingestion_metadata_task, adls_bulk_upload
from viadot.tasks.genesys import GenesysToCSV


@task(timeout=3600)
def add_timestamp(
    file_names: List = None,
    path: str = "",
    sep: str = "\t",
) -> None:
    """Add new column _viadot_downloaded_at_utc into every genesys file.

    Args:
        file_names (List, optional): All file names of downloaded files. Defaults to "\t".
        path (str, optional): Relative path to the file. Defaults to empty string.
        sep (str, optional): Separator in csv file. Defaults to None.
    """
    for file in file_names:
        df = pd.read_csv(os.path.join(path, file), sep=sep)
        df_updated = add_ingestion_metadata_task.run(df)
        df_updated.to_csv(os.path.join(path, file), index=False, sep=sep)


@task(timeout=3600)
def filter_userid(
    file_names: list = None,
    path: str = "",
    sep: str = "\t",
    user_ids: list = None,
    apply_method: bool = False,
) -> None:
    """Filter out the data frame by user ID.

    Args:
        file_names (list, optional): All file names of downloaded files. Defaults to "\t".
        path (str, optional): Relative path to the file. Defaults to empty string.
        user_ids (list, optional): List of all user IDs to select in the data frame. Defaults to None.
        sep (str, optional): Separator in csv file. Defaults to None.
        apply_method (bool, optional): Use this method or avoid its execution. Defaults to False.
    """

    if not file_names or not user_ids:
        apply_method = False

    if apply_method:
        for file in file_names:
            df = pd.read_csv(os.path.join(path, file), sep=sep)

            # first: it gets all the conversations ID where an agent is present.
            conversations_id = np.array([])
            for user in user_ids:
                user_filter = df["userId"] == user
                if any(user_filter):
                    ndf = df[user_filter]
                    conversations_id = np.append(
                        conversations_id, ndf["conversationId"]
                    )
            conversations_id = np.unique(conversations_id)

            # second: filter data frame out by the convesation id
            df2 = pd.DataFrame(columns=df.columns)
            for conversation in conversations_id:
                df_tmp = df[df["conversationId"] == conversation]
                df2 = pd.concat([df2, df_tmp])

            df2.to_csv(os.path.join(path, file), index=False, sep=sep)


class GenesysToADLS(Flow):
    def __init__(
        self,
        name: str,
        view_type: str = None,
        view_type_time_sleep: int = 80,
        post_data_list: List[str] = None,
        end_point: str = "analytics/reporting/exports",
        list_of_userids: list = None,
        start_date: str = None,
        end_date: str = None,
        sep: str = "\t",
        environment: str = None,
        report_url: str = None,
        report_columns: List[str] = None,
        conversationId_list: List[str] = None,
        key_list: List[str] = None,
        local_file_path: str = "",
        adls_file_path: str = None,
        overwrite_adls: bool = True,
        adls_sp_credentials_secret: str = None,
        credentials_genesys: Dict[str, Any] = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Genesys flow that generates CSV files and upload them to ADLS.

        Args:
            name (str): The name of the Flow.
            view_type (str, optional): The type of view export job to be created. Defaults to None.
            view_type_time_sleep (int, optional): Waiting time to retrieve data from Genesys API. Defaults to 80.
            post_data_list (List[str], optional): List of string templates to generate json body. Defaults to None.
                Example for only one POST:
                >>> post_data_list = '''[{
                >>>         "name": "AGENT_STATUS_DETAIL_VIEW",
                >>>         "timeZone": "UTC",
                >>>         "exportFormat": "CSV",
                >>>         "interval": ""2022-06-09T00:00:00/2022-06-10T00:00:00",
                >>>         "period": "PT30M",
                >>>         "viewType": "AGENT_STATUS_DETAIL_VIEW",
                >>>         "filter": {"userIds": ["aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"],},
                >>>         "read": True,
                >>>         "locale": "en-us",
                >>>         "hasFormatDurations": False,
                >>>         "hasSplitFilters": True,
                >>>         "excludeEmptyRows": True,
                >>>         "hasSummaryRow": False,
                >>>         "csvDelimiter": "COMMA",
                >>>         "hasCustomParticipantAttributes": True,
                >>>     }]'''
                If you need to add more POSTs in the same call, just add them to the list separated by a comma.
            end_point (str, optional): Final end point for Genesys connection. Defaults to "analytics/reporting/exports".
            list_of_userids (list, optional): List of all user IDs to select in the data frame. Defaults to None.
            start_date (str, optional): Start date of the report. Defaults to None.
            end_date (str, optional): End date of the report. Defaults to None.
            sep (str, optional): Separator in csv file. Defaults to "\t".
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
                from credentials.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.
            conversationId_list (List[str], optional): List of conversationId passed as attribute of GET method. Defaults to None.
            key_list (List[str], optional): List of keys needed to specify the columns in the GET request method. Defaults to None.
            local_file_path (str, optional): The local path from which to upload the file(s). Defaults to "".
            adls_file_path (str, optional): The destination path at ADLS. Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the data lake. Defaults to True.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            credentials(dict, optional): Credentials for the genesys api. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # GenesysToCSV
        self.flow_name = name
        self.view_type = view_type
        self.view_type_time_sleep = view_type_time_sleep
        self.post_data_list = post_data_list
        self.end_point = end_point
        if self.end_point == "analytics/conversations/details/query":
            self.apply_method = True
        else:
            self.apply_method = False
        self.list_of_userids = list_of_userids
        self.environment = environment
        self.report_url = report_url
        self.report_columns = report_columns
        self.conversationId_list = conversationId_list
        self.key_list = key_list
        self.start_date = start_date
        self.end_date = end_date
        self.sep = sep
        self.timeout = timeout

        # AzureDataLake
        self.local_file_path = local_file_path
        self.adls_file_path = adls_file_path
        self.overwrite_adls = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.credentials_genesys = credentials_genesys

        super().__init__(*args, name=self.flow_name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:
        to_csv = GenesysToCSV(
            timeout=self.timeout,
            local_file_path=self.local_file_path,
            sep=self.sep,
        )

        file_names = to_csv.bind(
            view_type=self.view_type,
            view_type_time_sleep=self.view_type_time_sleep,
            post_data_list=self.post_data_list,
            end_point=self.end_point,
            start_date=self.start_date,
            end_date=self.end_date,
            environment=self.environment,
            conversationId_list=self.conversationId_list,
            key_list=self.key_list,
            credentials_genesys=self.credentials_genesys,
            flow=self,
        )

        filter_userid.bind(
            file_names,
            path=self.local_file_path,
            sep=self.sep,
            user_ids=self.list_of_userids,
            apply_method=self.apply_method,
            flow=self,
        )

        add_timestamp.bind(
            file_names,
            path=self.local_file_path,
            sep=self.sep,
            flow=self,
        )

        adls_bulk_upload.bind(
            file_names=file_names,
            file_name_relative_path=self.local_file_path,
            adls_file_path=self.adls_file_path,
            adls_sp_credentials_secret=self.adls_sp_credentials_secret,
            timeout=self.timeout,
            flow=self,
        )

        filter_userid.set_upstream(file_names, flow=self)
        add_timestamp.set_upstream(filter_userid, flow=self)
        adls_bulk_upload.set_upstream(add_timestamp, flow=self)



import os
from unittest import mock

import pytest

from viadot.config import local_config
from viadot.sources import Mindful
from viadot.tasks import MindfulToCSV

os.system("clear")

credentials_mindful = local_config["MINDFUL"]
auth = (
    credentials_mindful["CUSTOMER_UUID"],
    credentials_mindful["AUTH_TOKEN"],
)


class MockClass:
    status_code = 200
    content = b'[{"id":7277599,"survey_id":505,"phone_number":"","survey_type":"inbound"},{"id":7277294,"survey_id":504,"phone_number":"","survey_type":"web"}]'

    def json():
        test = [
            {
                "id": 7277599,
                "survey_id": 505,
                "phone_number": "",
                "survey_type": "inbound",
            },
            {"id": 7277294, "survey_id": 504, "phone_number": "", "survey_type": "web"},
        ]
        return test


class MockClass2:
    status_code = 204
    content = b""

    def json():
        return None


@pytest.mark.init
def test_instance_mindful():
    mf = Mindful(auth=auth)
    assert isinstance(mf, Mindful)


@mock.patch("viadot.sources.mindful.handle_api_response", return_value=MockClass)
@pytest.mark.connect
def test_mindful_api_response(mock_connection):
    mf = Mindful(auth=auth)
    mf.get_interactions_list()
    mf.get_responses_list()
    mock_connection.call_count == 2


@mock.patch("viadot.sources.mindful.handle_api_response", return_value=MockClass)
@pytest.mark.connect
def test_mindful_api_response2(mock_api_response):
    mf = Mindful(auth=auth)

    response = mf.get_interactions_list()

    assert response.status_code == 200 and isinstance(response.json(), list)
    assert mf.endpoint == "interactions"


@mock.patch("viadot.sources.mindful.handle_api_response", return_value=MockClass)
@pytest.mark.connect
def test_mindful_api_response3(mock_api_response):
    mf = Mindful(auth=auth)

    response = mf.get_responses_list()

    assert response.status_code == 200 and isinstance(response.json(), list)
    assert mf.endpoint == "responses"


@mock.patch("viadot.sources.mindful.handle_api_response", return_value=MockClass)
@pytest.mark.connect
def test_mindful_api_response4(mock_api_response):
    mf = Mindful(auth=auth)

    response = mf.get_survey_list()

    assert response.status_code == 200 and isinstance(response.json(), list)
    assert mf.endpoint == "surveys"


@mock.patch("viadot.sources.Mindful._mindful_api_response", return_value=MockClass)
@pytest.mark.save
def test_mindful_interactions(mock_connection):
    mf = Mindful(auth=auth)
    response = mf.get_interactions_list()
    mf.response_to_file(response)
    assert mf.endpoint == "interactions" and isinstance(mf.endpoint, str)

    assert os.path.exists("interactions.csv")
    os.remove("interactions.csv")


@mock.patch("viadot.sources.Mindful._mindful_api_response", return_value=MockClass)
@pytest.mark.save
def test_mindful_responses(mock_connection):
    mf = Mindful(auth=auth)
    response = mf.get_responses_list()
    mf.response_to_file(response)

    assert mf.endpoint == "responses" and isinstance(mf.endpoint, str)
    assert os.path.exists("responses.csv")
    os.remove("responses.csv")


@mock.patch("viadot.sources.Mindful._mindful_api_response", return_value=MockClass)
@pytest.mark.save
def test_mindful_surveys(mock_connection):
    mf = Mindful(auth=auth)
    response = mf.get_survey_list()
    mf.response_to_file(response)

    assert mf.endpoint == "surveys" and isinstance(mf.endpoint, str)
    assert os.path.exists("surveys.csv")
    os.remove("surveys.csv")


@mock.patch("viadot.sources.Mindful._mindful_api_response", return_value=MockClass2)
@pytest.mark.exception
def test_file_exception(mock_mindful):
    mf = MindfulToCSV()
    response = mf.run(credentials_mindful=credentials_mindful)
    assert response == None



import os

import pytest
from prefect.engine.signals import FAIL

from viadot.tasks import BCPTask
from viadot.tasks.azure_sql import AzureSQLCreateTable, AzureSQLDBQuery

SCHEMA = "sandbox"
TABLE = "test_bcp"
FAIL_TABLE = "nonexistent_table"
ERROR_TABLE = "test_bcp_error"
ERROR_LOG_FILE = "log_file.log"


@pytest.fixture(scope="function")
def test_table():
    create_table_task = AzureSQLCreateTable()
    create_table_task.run(
        schema=SCHEMA,
        table=TABLE,
        dtypes={"country": "VARCHAR(25)", "sales": "INT"},
        if_exists="replace",
    )
    yield
    sql_query_task = AzureSQLDBQuery()
    sql_query_task.run(f"DROP TABLE {SCHEMA}.{TABLE};")


@pytest.fixture(scope="function")
def test_error_table():
    create_table_task = AzureSQLCreateTable()
    create_table_task.run(
        schema=SCHEMA,
        table=ERROR_TABLE,
        dtypes={"country": "INT", "sales": "INT"},
        if_exists="replace",
    )
    yield
    sql_query_task = AzureSQLDBQuery()
    sql_query_task.run(f"DROP TABLE  {SCHEMA}.{ERROR_TABLE};")


def test_bcp(TEST_CSV_FILE_PATH, test_table):
    bcp_task = BCPTask()

    try:
        result = bcp_task.run(
            path=TEST_CSV_FILE_PATH,
            schema=SCHEMA,
            table=TABLE,
            error_log_file_path=ERROR_LOG_FILE,
        )
    except FAIL:
        result = False

    assert result is not False
    os.remove(ERROR_LOG_FILE)


def test_bcp_fail(TEST_CSV_FILE_PATH, test_table):
    bcp_task = BCPTask()

    with pytest.raises(FAIL):
        bcp_task.run(path=TEST_CSV_FILE_PATH, schema=SCHEMA, table=FAIL_TABLE)


def test_bcp_log_error(TEST_CSV_FILE_PATH, test_error_table):
    bcp_task = BCPTask()
    bcp_task.run(
        path=TEST_CSV_FILE_PATH,
        schema=SCHEMA,
        table=ERROR_TABLE,
        error_log_file_path=ERROR_LOG_FILE,
        on_error="skip",
    )
    assert (
        os.path.exists(ERROR_LOG_FILE) is True and os.path.getsize(ERROR_LOG_FILE) != 0
    )
    os.remove(ERROR_LOG_FILE)



import os

import pendulum
from prefect.tasks.secrets import PrefectSecret

from viadot.flows import BigQueryToADLS
from viadot.tasks import AzureDataLakeRemove

ADLS_DIR_PATH = "raw/tests/"
ADLS_FILE_NAME = str(pendulum.now("utc")) + ".parquet"
BIGQ_CREDENTIAL_KEY = "BIGQUERY-TESTS"
ADLS_CREDENTIAL_SECRET = PrefectSecret(
    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
).run()


def test_bigquery_to_adls():
    flow_bigquery = BigQueryToADLS(
        name="Test BigQuery to ADLS extract",
        dataset_name="official_empty",
        table_name="space",
        adls_file_name=ADLS_FILE_NAME,
        credentials_key=BIGQ_CREDENTIAL_KEY,
        adls_dir_path=ADLS_DIR_PATH,
        adls_sp_credentials_secret=ADLS_CREDENTIAL_SECRET,
    )

    result = flow_bigquery.run()
    assert result.is_successful()

    task_results = result.result.values()
    assert all([task_result.is_successful() for task_result in task_results])

    os.remove("test_bigquery_to_adls_extract.parquet")
    os.remove("test_bigquery_to_adls_extract.json")


def test_bigquery_to_adls_overwrite_true():
    flow_bigquery = BigQueryToADLS(
        name="Test BigQuery to ADLS overwrite true",
        dataset_name="official_empty",
        table_name="space",
        credentials_key=BIGQ_CREDENTIAL_KEY,
        adls_file_name=ADLS_FILE_NAME,
        overwrite_adls=True,
        adls_dir_path=ADLS_DIR_PATH,
        adls_sp_credentials_secret=ADLS_CREDENTIAL_SECRET,
    )

    result = flow_bigquery.run()
    assert result.is_successful()

    task_results = result.result.values()
    assert all([task_result.is_successful() for task_result in task_results])
    os.remove("test_bigquery_to_adls_overwrite_true.parquet")
    os.remove("test_bigquery_to_adls_overwrite_true.json")


def test_bigquery_to_adls_false():
    flow_bigquery = BigQueryToADLS(
        name="Test BigQuery to ADLS overwrite false",
        dataset_name="official_empty",
        table_name="space",
        adls_file_name=ADLS_FILE_NAME,
        overwrite_adls=False,
        credentials_key=BIGQ_CREDENTIAL_KEY,
        adls_dir_path=ADLS_DIR_PATH,
        adls_sp_credentials_secret=ADLS_CREDENTIAL_SECRET,
    )

    result = flow_bigquery.run()
    assert result.is_failed()
    os.remove("test_bigquery_to_adls_overwrite_false.parquet")
    os.remove("test_bigquery_to_adls_overwrite_false.json")
    rm = AzureDataLakeRemove(
        path=ADLS_DIR_PATH + ADLS_FILE_NAME, vault_name="azuwevelcrkeyv001s"
    )
    rm.run(sp_credentials_secret=ADLS_CREDENTIAL_SECRET)



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import VidClubToADLS

DATA = {"col1": ["aaa", "bbb", "ccc"], "col2": [11, 22, 33]}
ADLS_FILE_NAME = "test_vid_club.parquet"
ADLS_DIR_PATH = "raw/test/"


@mock.patch(
    "viadot.tasks.VidClubToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_vidclub_to_adls_run_flow(mocked_class):
    flow = VidClubToADLS(
        "test_vidclub_to_adls_flow_run",
        source=["test"],
        from_date="2023-06-05",
        overwrite_adls=True,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_vidclub_to_adls_flow_run.parquet")
    os.remove("test_vidclub_to_adls_flow_run.json")



import os
import sqlite3

import pandas as pd
import pytest

from viadot.tasks.sqlite import SQLiteInsert, SQLiteQuery, SQLiteSQLtoDF

TABLE = "test"


@pytest.fixture(scope="session")
def sqlite_insert():
    task = SQLiteInsert()
    yield task


def test_sqlite_insert_proper(sqlite_insert):
    dtypes = {"AA": "INT", "BB": "INT"}
    df2 = pd.DataFrame({"AA": [1, 2, 3], "BB": [11, 22, 33]})
    sqlite_insert.run(
        table_name=TABLE,
        df=df2,
        dtypes=dtypes,
        if_exists="skip",
        db_path="testdb.sqlite",
    )

    with sqlite3.connect("testdb.sqlite") as db:
        cursor = db.cursor()
        cursor.execute("""SELECT COUNT(*) from test """)
        result = cursor.fetchall()
        assert result[0][0] != 0
    os.remove("testdb.sqlite")


def test_sqlite_insert_empty(caplog, sqlite_insert):
    df = pd.DataFrame()
    dtypes = {"AA": "INT", "BB": "INT"}
    sqlite_insert.run(
        table_name=TABLE,
        df=df,
        dtypes=dtypes,
        if_exists="skip",
        db_path="testdb1.sqlite",
    )

    with sqlite3.connect("testdb1.sqlite") as db:
        cursor2 = db.cursor()
        cursor2.execute("""SELECT COUNT(*) from test """)
        result2 = cursor2.fetchall()
        assert result2[0][0] == 0

    assert "DataFrame is empty" in caplog.text
    os.remove("testdb1.sqlite")


def test_sqlite_insert_not(caplog, sqlite_insert):
    dtypes = {"AA": "INT", "BB": "INT"}
    not_df = []
    sqlite_insert.run(
        table_name=TABLE,
        df=not_df,
        dtypes=dtypes,
        if_exists="skip",
        db_path="testdb2.sqlite",
    )
    with sqlite3.connect("testdb2.sqlite") as db:
        cursor3 = db.cursor()
        cursor3.execute("""SELECT COUNT(*) from test """)
        result3 = cursor3.fetchall()
        assert result3[0][0] == 0

    assert "Object is not a pandas DataFrame" in caplog.text
    os.remove("testdb2.sqlite")


def test_sqlite_sql_to_df(sqlite_insert):
    task = SQLiteSQLtoDF()
    with sqlite3.connect("testdb3.sqlite") as db:
        cursor = db.cursor()
        cursor.execute("""CREATE TABLE IF NOT EXISTS test ([AA] INT, [BB] INT) """)
        cursor.execute("""INSERT INTO test VALUES (11,22), (11,33)""")

    script = "SELECT * FROM test"
    with open("testscript.sql", "w") as file:
        file.write(script)
    df = task.run(db_path="testdb3.sqlite", sql_path="testscript.sql")
    assert isinstance(df, pd.DataFrame) == True
    expected = pd.DataFrame({"AA": [11, 11], "BB": [22, 33]})
    assert df.equals(expected) == True
    os.remove("testdb3.sqlite")
    os.remove("testscript.sql")


def test_sqlite_to_query(sqlite_insert):
    with sqlite3.connect("testdb4.sqlite") as db:
        cursor = db.cursor()
        cursor.execute("""CREATE TABLE IF NOT EXISTS test ([AA] INT, [BB] INT) """)
        query = "INSERT INTO test VALUES (11,22), (11,33)"
        task = SQLiteQuery()
        task.run(query=query, db_path="testdb4.sqlite")
        cursor.execute("""SELECT COUNT(*) from test """)
        result = cursor.fetchall()
        assert result[0][0] != 0
    os.remove("testdb4.sqlite")



import json
import urllib
from copy import deepcopy
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from prefect.utilities import logging

from ..config import local_config
from ..exceptions import CredentialError
from ..utils import handle_api_response
from .base import Source

logger = logging.get_logger(__name__)


class Supermetrics(Source):
    """
    A class implementing the Supermetrics API.

    Documentation for this API is located at: https://supermetrics.com/docs/product-api-getting-started/
    Usage limits: https://supermetrics.com/docs/product-api-usage-limits/

    Parameters
    ----------
    query_params : Dict[str, Any], optional
        The parameters to pass to the GET query.
        See https://supermetrics.com/docs/product-api-get-data/ for full specification,
        by default None
    """

    API_ENDPOINT = "https://api.supermetrics.com/enterprise/v2/query/data/json"

    def __init__(self, *args, query_params: Dict[str, Any] = None, **kwargs):
        DEFAULT_CREDENTIALS = local_config.get("SUPERMETRICS")
        credentials = kwargs.pop("credentials", DEFAULT_CREDENTIALS)
        if credentials is None:
            raise CredentialError("Missing credentials.")
        super().__init__(*args, credentials=credentials, **kwargs)
        self.query_params = query_params

    @classmethod
    def get_params_from_api_query(cls, url: str) -> Dict[str, Any]:
        """Returns parmeters from API query in a dictionary"""
        url_unquoted = urllib.parse.unquote(url)
        s = urllib.parse.parse_qs(url_unquoted)
        endpoint = list(s.keys())[0]
        params = s[endpoint][0]
        params_d = json.loads(params)
        return params_d

    @classmethod
    def from_url(cls, url: str, credentials: Dict[str, Any] = None):
        obj = Supermetrics(credentials=credentials or local_config.get("SUPERMETRICS"))
        params = cls.get_params_from_api_query(url)
        obj.query_params = params
        return obj

    def to_json(self, timeout=(3.05, 60 * 30)) -> Dict[str, Any]:
        """Download query results to a dictionary.
        Note that Supermetrics API will sometimes hang and not return any error message,
        so we're adding a timeout to GET.

        See [requests docs](https://docs.python-requests.org/en/master/user/advanced/#timeouts)
        for an explanation of why this timeout value will work on long-running queries but fail fast
        on connection issues.
        """

        if not self.query_params:
            raise ValueError("Please build the query first")

        params = {"json": json.dumps(self.query_params)}
        headers = {"Authorization": f'Bearer {self.credentials["API_KEY"]}'}

        response = handle_api_response(
            url=self.API_ENDPOINT, params=params, headers=headers, timeout=timeout
        )
        return response.json()

    @classmethod
    def _get_col_names_google_analytics(
        cls,
        response: dict,
    ) -> List[str]:
        """Returns list of Google Analytics columns names"""

        # Supermetrics allows pivoting GA data, in which case it generates additional columns,
        # which are not enlisted in response's query metadata but are instead added as the first row of data.
        is_pivoted = any(
            field["field_split"] == "column"
            for field in response["meta"]["query"]["fields"]
        )

        if is_pivoted:
            if not response["data"]:
                raise ValueError(
                    "Couldn't find column names as query returned no data."
                )
            columns = response["data"][0]
        else:
            # non-pivoted data; query fields match result fields
            cols_meta = response["meta"]["query"]["fields"]
            columns = [col_meta["field_name"] for col_meta in cols_meta]
        return columns

    @classmethod
    def _get_col_names_other(cls, response: dict) -> List[str]:
        """Returns list of columns names (to Google Analytics use  _get_col_names_google_analytics ()"""
        cols_meta = response["meta"]["query"]["fields"]
        columns = [col_meta["field_name"] for col_meta in cols_meta]
        return columns

    def _get_col_names(self) -> List[str]:
        """Returns list of columns names"""

        query_params_cp = deepcopy(self.query_params)
        query_params_cp["offset_start"] = 0
        query_params_cp["offset_end"] = 0
        response: dict = Supermetrics(query_params=query_params_cp).to_json()
        if self.query_params["ds_id"] == "GA":
            return Supermetrics._get_col_names_google_analytics(response)
        else:
            return Supermetrics._get_col_names_other(response)

    def to_df(self, if_empty: str = "warn") -> pd.DataFrame:
        """Download data into a pandas DataFrame.

        Note that Supermetric can calculate some fields on the fly and alias them in the
        returned result. For example, if the query requests the `position` field,
        Supermetric may return an `Average position` caclulated field.
        For this reason we take columns names from the actual results rather than from input fields.

        Args:
            if_empty (str, optional): What to do if query returned no data. Defaults to "warn".

        Returns:
            pd.DataFrame: the DataFrame containing query results
        """
        try:
            columns = self._get_col_names()
        except ValueError:
            columns = None

        data = self.to_json()["data"]

        if data:
            df = pd.DataFrame(data[1:], columns=columns).replace("", np.nan)
        else:
            df = pd.DataFrame(columns=columns)

        if df.empty:
            self._handle_if_empty(if_empty)

        return df

    def query(self, params: Dict[str, Any]):
        self.query_params = params
        self.query_params["api_key"] = self.credentials["API_KEY"]
        return self



from datetime import datetime, timedelta
from unittest import mock

import pandas as pd
import pytest

from viadot.exceptions import ValidationError
from viadot.sources import VidClub
from viadot.task_utils import credentials_loader

CREDENTIALS = credentials_loader.run(credentials_secret="VIDCLUB")
vc = VidClub(credentials=CREDENTIALS)


@pytest.fixture
def var_dictionary():
    variables = {}

    return variables


class MockClass:
    status_code = 200

    def json():
        df = pd.DataFrame()
        return df


@pytest.mark.init
def test_default_credential_param():
    """
    Checks if credentials are loaded from Azure Key Vault or PrefectSecret or from local config ursing credentials_loader and if it's dictionary type.
    """
    assert vc.credentials is not None and isinstance(vc.credentials, dict)


@pytest.mark.build_query
def test_build_query_wrong_source():
    """
    Checks if passing different source than Literal["jobs", "product", "company", "survey"] is catched and returns error.
    """
    with pytest.raises(
        ValidationError, match=r"Pick one these sources: jobs, product, company, survey"
    ):
        query = vc.build_query(
            source="test",
            from_date="2023-03-24",
            to_date="2023-03-24",
            api_url="test",
            items_per_page=1,
        )


@pytest.mark.build_query
def test_url_string():
    """
    Checks if fucntion generates URL with needed parameters.
    """
    source = "jobs"
    from_date = "2023-03-24"
    to_date = "2023-03-24"
    api_url = "https://api/test/"
    items_per_page = 1

    expected_elements = [
        f"from={from_date}",
        f"to={to_date}",
        "region=all",
        f"limit={items_per_page}",
        api_url,
    ]

    query = vc.build_query(
        source=source,
        from_date=from_date,
        to_date=to_date,
        api_url=api_url,
        items_per_page=items_per_page,
    )

    assert all(ex in query for ex in expected_elements)


@pytest.mark.intervals
def test_intervals_split():
    """
    Checks if prrovided date range with days_interval creates list with expected split.
    """
    from_date = "2022-01-01"
    to_date = "2022-01-19"
    days_interval = 5
    expected_starts = ["2022-01-01", "2022-01-06", "2022-01-11", "2022-01-16"]
    expected_ends = ["2022-01-06", "2022-01-11", "2022-01-16", "2022-01-19"]
    starts, ends = vc.intervals(
        from_date=from_date, to_date=to_date, days_interval=days_interval
    )

    assert starts == expected_starts
    assert ends == expected_ends


@pytest.mark.connection_check
def test_check_connection():
    """
    Checks if check_connection method returns tuple with dictionary and string.
    """
    output = vc.check_connection(
        source="jobs",
        from_date="2023-03-24",
        to_date="2023-03-24",
        items_per_page=1,
    )

    response, first_url = vc.check_connection(
        source="jobs",
        from_date="2023-03-24",
        to_date="2023-03-24",
        items_per_page=1,
    )

    assert isinstance(output, tuple)
    assert isinstance(response, dict)
    assert isinstance(first_url, str)


@pytest.mark.proper
def test_get_response_wrong_source():
    """
    Checks if ValidationError is returned when passing wrong source name.
    """
    with pytest.raises(
        ValidationError, match=r"The source has to be: jobs, product, company or survey"
    ):
        query = vc.get_response(source="test")


@mock.patch(
    "viadot.sources.vid_club.VidClub.get_response", return_value=MockClass.json()
)
@pytest.mark.parametrize("source", ["jobs", "company", "product", "survey"])
@pytest.mark.proper
def test_get_response_sources(mock_api_response, source):
    """
    Checks if get_response method returnes DataFrame for each of the 4 possible sources.
    Test assert that the mock was called exactly once.

    Args:
        mock_api_response: Mocked return_value for get_response method.
        source: The endpoint source to be accessed.
    """
    query = vc.get_response(source=source, to_date="2023-03-24", from_date="2023-03-24")

    assert isinstance(query, pd.DataFrame)
    mock_api_response.assert_called_once()


@pytest.mark.proper
def test_get_response_wrong_date():
    """
    Checks if ValidationError is returned when passing from_date earlier than 2022-03-22.
    """
    with pytest.raises(
        ValidationError, match=r"from_date cannot be earlier than 2022-03-22"
    ):
        vc.get_response(source="jobs", from_date="2021-05-09")


@pytest.mark.proper
def test_get_response_wrong_date_range():
    """
    Checks if ValidationError is returned when passing to_date earlier than from_date.
    """
    with pytest.raises(
        ValidationError, match=r"to_date cannot be earlier than from_date"
    ):
        vc.get_response(source="jobs", to_date="2022-05-04", from_date="2022-05-05")


@pytest.mark.total_load
def test_total_load_for_the_same_dates():
    """
    total_load method includes logic for situation when from_date == to_date. In this scenario interval split is skipped and used just get_response method.
    This test checks if this logic is executed without error and returned object if DataFrame.
    """
    from_date = "2022-04-01"
    to_date = "2022-04-01"
    days_interval = 10
    source = "jobs"
    df = vc.total_load(
        from_date=from_date, to_date=to_date, days_interval=days_interval, source=source
    )

    assert isinstance(df, pd.DataFrame)


@pytest.mark.total_load
def test_total_load_for_intervals():
    """
    Checks if interval function is properly looped in the total_load method. At first we check if returned object is DataFrame,
    then we check if returned DataFrame for smaller date range contains less rows than DataFrame returned for bigger date range.
    """
    from_date = "2022-04-01"
    to_date = "2022-04-12"
    days_interval = 2
    source = "jobs"

    date_object = datetime.strptime(from_date, "%Y-%m-%d") + timedelta(
        days=days_interval
    )
    one_interval = date_object.strftime("%Y-%m-%d")

    df = vc.total_load(
        from_date=from_date, to_date=to_date, days_interval=days_interval, source=source
    )
    df_one_interval = vc.total_load(
        from_date=from_date,
        to_date=one_interval,
        days_interval=days_interval,
        source=source,
    )

    assert isinstance(df, pd.DataFrame)
    assert len(df) > len(df_one_interval)


@pytest.mark.total_load
def test_drop_duplicates():
    """
    Checks logic for dropping duplicated rows, that is included in total_load method.
    Test checks if returned DataFrame has duplicates.
    """
    from_date = "2022-04-01"
    to_date = "2022-04-12"
    days_interval = 2
    source = "jobs"

    df = vc.total_load(
        from_date=from_date, to_date=to_date, days_interval=days_interval, source=source
    )
    dups_mask = df.duplicated()
    df_check = df[dups_mask]

    assert len(df_check) == 0



import os
from io import StringIO
from datetime import datetime, timedelta
from typing import Any, Dict, Literal, Tuple

import pandas as pd
import prefect
from requests.models import Response
from requests.auth import HTTPBasicAuth

from viadot.exceptions import APIError
from viadot.sources.base import Source
from viadot.utils import handle_api_response


class Mindful(Source):
    def __init__(
        self,
        auth: Tuple[str],
        region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
        start_date: datetime = None,
        end_date: datetime = None,
        date_interval: int = 1,
        file_extension: Literal["parquet", "csv"] = "csv",
        *args,
        **kwargs,
    ) -> None:
        """Mindful connector which allows listing and downloading into Data Frame or specified format output.

        Args:
            auth (Tuple[str]): Authentication credentials for calling Mindful API. The structure is user and password.
            region (Literal[us1, us2, us3, ca1, eu1, au1], optional): SD region from where to interact with the mindful API. Defaults to "eu1".
            start_date (datetime, optional): Start date of the request. Defaults to None.
            end_date (datetime, optional): End date of the resquest. Defaults to None.
            date_interval (int, optional): How many days are included in the request.
                If end_date is passed as an argument, date_interval will be invalidated. Defaults to 1.
            file_extension (Literal[parquet, csv], optional): File extensions for storing responses. Defaults to "csv".
        """
        self.logger = prefect.context.get("logger")

        super().__init__(*args, **kwargs)

        if region != "us1":
            self.region = region + "."
        else:
            self.region = ""

        if not isinstance(start_date, datetime):
            self.start_date = datetime.now() - timedelta(days=date_interval)
            self.end_date = self.start_date + timedelta(days=date_interval)
            self.logger.info(
                f"Mindful start_date variable is None or not in datetime format, it has been taken as: {self.start_date}."
            )
            self.logger.info(
                f"Mindful end_date variable has been also taken as: {self.end_date}."
            )
        elif isinstance(start_date, datetime) and not isinstance(end_date, datetime):
            self.start_date = start_date
            self.end_date = start_date + timedelta(days=date_interval)
            if self.end_date > datetime.now():
                self.end_date = datetime.now()
            self.logger.info(
                f"Mindful end_date variable is None or not in datetime format, it has been taken as: {self.end_date}."
            )
        elif start_date >= end_date:
            raise ValueError(
                f"start_date variable must be lower than end_date variable."
            )
        else:
            self.start_date = start_date
            self.end_date = end_date
            self.logger.info(
                f"Mindful files to download will store data from {self.start_date} to {self.end_date}."
            )

        self.file_extension = file_extension
        self.auth = auth

    def _mindful_api_response(
        self,
        params: Dict[str, Any] = None,
        endpoint: str = "",
        **kwargs,
    ) -> Response:
        """Basic call to Mindful API given an endpoint.

        Args:
            params (Dict[str, Any], optional): Parameters to be passed into the request. Defaults to None.
            endpoint (str, optional): API endpoint for an individual request. Defaults to "".

        Returns:
            Response: request object with the response from the Mindful API.
        """

        response = handle_api_response(
            url=f"https://{self.region}surveydynamix.com/api/{endpoint}",
            params=params,
            method="GET",
            auth=HTTPBasicAuth(*self.auth),
        )

        return response

    def get_interactions_list(
        self,
        limit: int = 1000,
        **kwargs,
    ) -> Response:
        """Gets a list of survey interactions as a JSON array of interaction resources.

        Args:
            limit (int, optional): The number of matching interactions to return. Defaults to 1000.

        Returns:
            Response: request object with the response from the Mindful API.
        """

        self.endpoint = "interactions"
        params = {
            "_limit": limit,
            "start_date": f"{self.start_date}",
            "end_date": f"{self.end_date}",
        }

        response = self._mindful_api_response(
            endpoint=self.endpoint,
            params=params,
        )

        if response.status_code == 200:
            self.logger.info(
                "Succesfully downloaded interactions data from mindful API."
            )
        elif response.status_code == 204 and not response.content.decode():
            self.logger.warning(
                f"Thera are not interactions data to download from {self.start_date} to {self.end_date}."
            )
        else:
            self.logger.error(
                f"Failed to downloaded interactions data. - {response.content}"
            )
            raise APIError("Failed to downloaded interactions data.")

        return response

    def get_responses_list(
        self,
        limit: int = 1000,
        **kwargs,
    ) -> Response:
        """Gets a list of survey responses associated with a survey, question, or interaction resource.

        Args:
            limit (int, optional): The number of matching interactions to return. Defaults to 1000.

        Returns:
            Response: request object with the response from the Mindful API.
        """

        self.endpoint = "responses"
        params = {
            "_limit": limit,
            "start_date": f"{self.start_date}",
            "end_date": f"{self.end_date}",
        }

        response = self._mindful_api_response(
            endpoint=self.endpoint,
            params=params,
        )

        if response.status_code == 200:
            self.logger.info("Succesfully downloaded responses data from mindful API.")
        elif response.status_code == 204 and not response.content.decode():
            self.logger.warning(
                f"Thera are not responses data to download from {self.start_date} to {self.end_date}."
            )
        else:
            self.logger.error(
                f"Failed to downloaded responses data. - {response.content}"
            )
            raise APIError("Failed to downloaded responses data.")

        return response

    def get_survey_list(
        self,
        limit: int = 1000,
        **kwargs,
    ) -> Response:
        """Gets a list of survey resources associated with the authenticated customer.

        Args:
            limit (int, optional): The number of matching interactions to return. Defaults to 1000.

        Returns:
            Response: Request object with the response from the Mindful API.
        """
        self.endpoint = "surveys"
        params = {
            "_limit": limit,
        }

        response = self._mindful_api_response(
            endpoint=self.endpoint,
            params=params,
        )

        if response.status_code == 200:
            self.logger.info("Succesfully downloaded surveys data from mindful API.")
        elif response.status_code == 204 and not response.content.decode():
            self.logger.warning(
                f"Thera are not surveys data to download from {self.start_date} to {self.end_date}."
            )
        else:
            self.logger.error(f"Failed to download surveys data. - {response.content}")
            raise APIError("Failed to downloaded surveys data.")

        return response

    def response_to_file(
        self,
        response: Response,
        file_name: str = None,
        file_path: str = "",
        sep: str = "\t",
    ) -> str:
        """Save Mindful response data to file and return filename.

        Args:
            response (Response): Request object with the response from the Mindful API.
            file_name (str, optional): Name of the file where saving data. Defaults to None.
            file_path (str, optional): Path where to save the file locally. Defaults to ''.
            sep (str, optional): Separator in csv file. Defaults to "\t".

        returns
            str: the absolute path of the downloaded file.
        """

        data_frame = pd.read_json(StringIO(response.content.decode("utf-8")))
        if file_name is None:
            complete_file_name = f"{self.endpoint}.{self.file_extension}"
            relative_path = os.path.join(file_path, complete_file_name)
        else:
            complete_file_name = f"{file_name}.{self.file_extension}"
            relative_path = os.path.join(file_path, complete_file_name)

        if self.file_extension == "csv":
            data_frame.to_csv(relative_path, index=False, sep=sep)
        elif self.file_extension == "parquet":
            data_frame.to_parquet(relative_path, index=False)
        else:
            self.logger.warning(
                "File extension is not available, please choose file_extension: 'parquet' or 'csv' (def.) at Mindful instance."
            )

        return relative_path



from typing import Any, Dict, List

from prefect import Flow, task
from prefect.tasks.prefect import create_flow_run, wait_for_flow_run
from prefect.utilities import logging

logger = logging.get_logger()


@task(timeout=3600)
def run_flows_list(flow_name: str, flows_list: List[List] = [List[None]]):
    """
    Task for running multiple flows in the given order. Task will create flow of flows.
    Args:
        flow_name(str): Name of a new flow.
        flows_list(List[List]): List containing lists of flow names and project names - [["flow1_name" , "project_name"], ["flow2_name" , "project_name"]].
            Flows have to be in the correct oreder. Defaults to [List[None]].
    """
    with Flow(flow_name) as flow:
        for i in range(len(flows_list)):
            exec(
                f"flow_{i}= create_flow_run(flow_name=flows_list[i][0], project_name=flows_list[i][1])"
            )
            exec(
                f"wait_for_flow_{i} = wait_for_flow_run(flow_{i}, raise_final_state=True)"
            )
        for i in range(1, len(flows_list)):
            exec(f"flow_{i}.set_upstream(wait_for_flow_{i-1})")
        flow_state = flow.run()
        if flow_state.is_failed():
            logger.error("One of the flows has failed!")
            raise ValueError("One of the flows has failed!")
        else:
            logger.info("All of the tasks succeeded.")


class MultipleFlows(Flow):
    """Flow to run multiple flows in the given order.
    Args:
        flow_name(str): Name of a new flow.
        flows_list(List[List]): List containing lists of flow names and project names - [["flow1_name" , "project_name"], ["flow2_name" , "project_name"]].
            Flows have to be in the correct oreder. Defaults to [List[None]].
    """

    def __init__(
        self,
        name: str,
        flows_list: List[List] = [List[None]],
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        self.flows_list = flows_list
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    def gen_flow(self) -> Flow:
        run_flows_list.bind(
            flow_name=self.name,
            flows_list=self.flows_list,
            flow=self,
        )



from viadot.config import local_config
from viadot.tasks import EpicorOrdersToDF


def test_epicor_orders_to_df():
    task = EpicorOrdersToDF(
        config_key="EPICOR",
        base_url=local_config.get("EPICOR").get("test_url"),
        filters_xml="""
<OrderQuery>
    <QueryFields>
        <CompanyNumber>001</CompanyNumber>
        <BegInvoiceDate>2022-05-16</BegInvoiceDate>
        <EndInvoiceDate>2022-05-16</EndInvoiceDate>
        <RecordCount>3</RecordCount>
    </QueryFields>
</OrderQuery>""",
    )
    df = task.run()
    assert df.shape == (24, 60)



from typing import Any, Dict, List, Literal, OrderedDict

import pandas as pd
from prefect.utilities import logging
from simple_salesforce import Salesforce as SF
from simple_salesforce.exceptions import SalesforceMalformedRequest

from ..config import local_config
from ..exceptions import CredentialError
from .base import Source

logger = logging.get_logger(__name__)


class Salesforce(Source):
    """
    A class for pulling data from theSalesforce.
    Args:
        domain (str): domain of a connection; defaults to 'test' (sandbox). Can be added only if built-in username/password/security token is provided.
        client_id (str): client id to keep the track of API calls.
        credentials (dict): credentials to connect with. If not provided, will read from local config file.
        env (Literal): environment information, provides information about credential and connection configuration; defaults to 'DEV'.
    ----------
    """

    def __init__(
        self,
        *args,
        domain: str = "test",
        client_id: str = "viadot",
        credentials: Dict[str, Any] = None,
        env: Literal["DEV", "QA", "PROD"] = "DEV",
        **kwargs,
    ):
        try:
            DEFAULT_CREDENTIALS = local_config["SALESFORCE"].get(env)
        except KeyError:
            DEFAULT_CREDENTIALS = None

        self.credentials = credentials or DEFAULT_CREDENTIALS or {}

        if self.credentials is None:
            raise CredentialError("Credentials not found.")

        super().__init__(*args, credentials=self.credentials, **kwargs)

        if env.upper() == "DEV":
            self.salesforce = SF(
                username=self.credentials["username"],
                password=self.credentials["password"],
                security_token="",
                domain=domain,
                client_id=client_id,
            )
        elif env.upper() == "QA":
            self.salesforce = SF(
                username=self.credentials["username"],
                password=self.credentials["password"],
                security_token=self.credentials["token"],
                domain=domain,
                client_id=client_id,
            )
        elif env.upper() == "PROD":
            self.salesforce = SF(
                username=self.credentials["username"],
                password=self.credentials["password"],
                security_token=self.credentials["token"],
                domain=domain,
                client_id=client_id,
            )
        else:
            raise ValueError("The only available environments are DEV, QA, and PROD.")

    def upsert(
        self,
        df: pd.DataFrame,
        table: str,
        external_id: str = None,
        raise_on_error: bool = False,
    ) -> None:
        if df.empty:
            logger.info("No data to upsert.")
            return

        if external_id and external_id not in df.columns:
            raise ValueError(
                f"Passed DataFrame does not contain column '{external_id}'."
            )

        table_to_upsert = getattr(self.salesforce, table)
        records = df.to_dict("records")
        records_cp = records.copy()

        for record in records_cp:
            response = 0
            if external_id:
                if record[external_id] is None:
                    continue
                else:
                    merge_key = f"{external_id}/{record[external_id]}"
                    record.pop(external_id)
            else:
                merge_key = record.pop("Id")

            try:
                response = table_to_upsert.upsert(data=record, record_id=merge_key)
            except SalesforceMalformedRequest as e:
                msg = f"Upsert of record {merge_key} failed."
                if raise_on_error:
                    raise ValueError(msg) from e
                else:
                    self.logger.warning(msg)

            codes = {200: "updated", 201: "created", 204: "updated"}

            if response not in codes:
                msg = f"Upsert failed for record: \n{record} with response {response}"
                if raise_on_error:
                    raise ValueError(msg)
                else:
                    self.logger.warning(msg)
            else:
                logger.info(f"Successfully {codes[response]} record {merge_key}.")

        logger.info(
            f"Successfully upserted {len(records)} records into table '{table}'."
        )

    def bulk_upsert(
        self,
        df: pd.DataFrame,
        table: str,
        external_id: str = None,
        batch_size: int = 10000,
        raise_on_error: bool = False,
    ) -> None:
        if df.empty:
            logger.info("No data to upsert.")
            return

        if external_id and external_id not in df.columns:
            raise ValueError(
                f"Passed DataFrame does not contain column '{external_id}'."
            )
        records = df.to_dict("records")
        response = 0
        try:
            response = self.salesforce.bulk.__getattr__(table).upsert(
                data=records, external_id_field=external_id, batch_size=batch_size
            )
        except SalesforceMalformedRequest as e:
            # Bulk insert didn't work at all.
            raise ValueError(f"Upsert of records failed: {e}") from e

        logger.info(f"Successfully upserted bulk records.")

        if any(result.get("success") is not True for result in response):
            # Upsert of some individual records failed.
            failed_records = [
                result for result in response if result.get("success") is not True
            ]
            msg = f"Upsert failed for records {failed_records} with response {response}"
            if raise_on_error:
                raise ValueError(msg)
            else:
                self.logger.warning(msg)

        logger.info(
            f"Successfully upserted {len(records)} records into table '{table}'."
        )

    def download(
        self, query: str = None, table: str = None, columns: List[str] = None
    ) -> List[OrderedDict]:
        if not query:
            if columns:
                columns_str = ", ".join(columns)
            else:
                columns_str = "FIELDS(STANDARD)"
            query = f"SELECT {columns_str} FROM {table}"
        records = self.salesforce.query(query).get("records")
        # Take trash out.
        _ = [record.pop("attributes") for record in records]
        return records

    def to_df(
        self,
        query: str = None,
        table: str = None,
        columns: List[str] = None,
        if_empty: str = None,
    ) -> pd.DataFrame:
        # TODO: handle if_empty, add typing (should be Literal)
        records = self.download(query=query, table=table, columns=columns)

        if not records:
            raise ValueError(f"Query produced no data.")

        return pd.DataFrame(records)



from viadot.config import local_config
from viadot.tasks import SAPRFCToDF


def test_sap_rfc_to_df_bbp():
    sap_test_creds = local_config.get("SAP").get("QA")
    task = SAPRFCToDF(
        credentials=sap_test_creds,
        query="SELECT MATNR, MATKL, MTART, LAEDA FROM MARA WHERE LAEDA LIKE '2022%'",
        func="BBP_RFC_READ_TABLE",
    )
    df = task.run()
    assert len(df.columns) == 4 and not df.empty



import os

from viadot.tasks.github import DownloadGitHubFile

# def test_github_clone_task():
#     clone_repo_task = CloneRepo()
#     repo = "fishtown-analytics/dbt"
#     repo_name = repo.split("/")[-1]
#     clone_repo_task.run(repo=repo)
#     assert os.path.exists(repo_name)


def test_download_github_file():
    task = DownloadGitHubFile()
    task.run(repo="dyvenia/viadot", from_path="LICENSE")
    assert os.path.exists("LICENSE")
    os.remove("LICENSE")



import shutil
from typing import Dict

import pygit2
from prefect.utilities.logging import get_logger
from prefect.utilities.tasks import defaults_from_attrs

from viadot.task_utils import *


class CloneRepo(Task):
    """Tasks for interacting with git repositories."""

    def __init__(
        self,
        url: str,
        token: str = None,
        token_secret: str = None,
        timeout: int = 60 * 10,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Clone Azure DevOps or GitHub repository.

        Args:
            url (str): Alternatively to all above, you can also provide the full URL (including token, if the repo is private) here.
                Examples:
                    Azure DevOps (public repo): https://dev.azure.com/{organization_name}/{project_name}/_git/{repo_name}  # noqa
                    Azure DevOps (private repo): https://{token}@dev.azure.com/{organization_name}/{project_name}/_git/{repo_name}  # noqa
                    GitHub (public repo): https://github.com/{organization_name}/{repo_name}.git  # noqa
                    GitHub (private repo): https://{token}@github.com/{organization_name}/{repo_name}.git  # noqa
            token (str, optional): The personal access token. Defaults to None.
            token_secret (str, optional): The name of the secret storing the token. Defaults to None.
                Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before a timeout occurs.
                Defaults to 600.
            kwargs (dict): Keyword arguments to be passed to `pygit2.clone_repository()`.
        """

        self.url = url
        self.token = token
        self.token_secret = token_secret

        super().__init__(
            name="clone_repo",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Clone repository"""
        super().__call__(self)

    @defaults_from_attrs("url", "token", "token_secret")
    def run(
        self,
        url: str,
        token: str = None,
        token_secret: str = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        """Run methon of CloneRepo task.

        Args:
            url (str): Alternatively to all above, you can also provide the full URL (including token, if the repo is private) here.
                Examples:
                    Azure DevOps (public repo): https://dev.azure.com/{organization_name}/{project_name}/_git/{repo_name}  # noqa
                    Azure DevOps (private repo): https://{token}@dev.azure.com/{organization_name}/{project_name}/_git/{repo_name}  # noqa
                    GitHub (public repo): https://github.com/{organization_name}/{repo_name}.git  # noqa
                    GitHub (private repo): https://{token}@github.com/{organization_name}/{repo_name}.git  # noqa
            token (str, optional): The personal access token. Defaults to None.
            token_secret (str, optional): The name of the secret storing the token. Defaults to None.
                Defaults to None.
        """
        logger = get_logger(__name__)

        url = url.strip("/")

        azure_secret_task = AzureKeyVaultSecret()
        token = token or azure_secret_task.run(secret=token_secret)
        if token:
            url = url.replace("https://dev.azure.com", f"https://{token}@dev.azure.com")
            url = url.replace("https://github.com", f"https://{token}@github.com")

        repo_name = url.split("/")[-1].replace(".git", "")
        path = kwargs.get("path") or repo_name
        kwargs["path"] = path

        logger.info(f"Removing {path}...")
        shutil.rmtree(path, ignore_errors=True)  # Delete folder on run

        logger.info(f"Cloning repo '{repo_name}' into {path}...")
        pygit2.clone_repository(url, **kwargs)
        logger.info(f"Repo '{repo_name}' has been successfully cloned into {path}.")



from operator import ne
from typing import Any, Dict, List, Literal

from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

logger = logging.get_logger()

from viadot.task_utils import (
    add_ingestion_metadata_task,
    cast_df_to_str,
    df_to_parquet,
    set_new_kv,
)
from viadot.tasks import DuckDBCreateTableFromParquet, SAPRFCToDF


class SAPToDuckDB(Flow):
    def __init__(
        self,
        query: str,
        table: str,
        local_file_path: str,
        func: str = "RFC_READ_TABLE",
        rfc_total_col_width_character_limit: int = 400,
        name: str = None,
        sep: str = None,
        schema: str = None,
        table_if_exists: Literal[
            "fail", "replace", "append", "skip", "delete"
        ] = "fail",
        if_empty: Literal["warn", "skip", "fail"] = "skip",
        sap_credentials: dict = None,
        duckdb_credentials: dict = None,
        update_kv: bool = False,
        filter_column: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """A flow for moving data from SAP to DuckDB.

        Args:
            query (str): The query to be executed on SAP with pyRFC.
            table (str): Destination table in DuckDB.
            local_file_path (str): The path to the source Parquet file.
            func (str, optional): SAP RFC function to use. Defaults to "RFC_READ_TABLE".
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in
            chunks in case of too many columns for RFC function. According to SAP documentation, the
            limit is 512 characters. However, we observed SAP raising an exception even on a slightly
            lower number of characters, so we add a safety margin. Defaults to 400.
            name (str, optional): The name of the flow. Defaults to None.
            sep (str, optional): The separator to use when reading query results. If not provided,
            multiple options are automatically tried. Defaults to None.
            schema (str, optional): Destination schema in DuckDB. Defaults to None.
            table_if_exists (Literal, optional):  What to do if the table already exists. Defaults to "fail".
            if_empty (Literal, optional): What to do if Parquet file is empty. Defaults to "skip".
            sap_credentials (dict, optional): The credentials to use to authenticate with SAP.
            By default, they're taken from the local viadot config.
            duckdb_credentials (dict, optional): The config to use for connecting with DuckDB. Defaults to None.
            update_kv (bool, optional): Whether or not to update key value on Prefect. Defaults to False.
            filter_column (str, optional): Name of the field based on which key value will be updated. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        # SAPRFCToDF
        self.query = query
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit
        self.sep = sep
        self.sap_credentials = sap_credentials

        # DuckDBCreateTableFromParquet
        self.table = table
        self.schema = schema
        self.if_exists = table_if_exists
        self.if_empty = if_empty
        self.local_file_path = local_file_path or self.slugify(name) + ".parquet"
        self.duckdb_credentials = duckdb_credentials
        self.update_kv = update_kv
        self.filter_column = filter_column

        super().__init__(*args, name=name, **kwargs)

        self.sap_to_df_task = SAPRFCToDF(credentials=sap_credentials, timeout=timeout)
        self.create_duckdb_table_task = DuckDBCreateTableFromParquet(
            credentials=duckdb_credentials, timeout=timeout
        )

        self.gen_flow()

    def gen_flow(self) -> Flow:
        df = self.sap_to_df_task.bind(
            query=self.query,
            sep=self.sep,
            func=self.func,
            rfc_total_col_width_character_limit=self.rfc_total_col_width_character_limit,
            flow=self,
        )

        df_mapped = cast_df_to_str.bind(df, flow=self)
        df_with_metadata = add_ingestion_metadata_task.bind(df_mapped, flow=self)
        parquet = df_to_parquet.bind(
            df=df_with_metadata,
            path=self.local_file_path,
            if_exists=self.if_exists,
            flow=self,
        )

        table = self.create_duckdb_table_task.bind(
            path=self.local_file_path,
            schema=self.schema,
            table=self.table,
            if_exists=self.if_exists,
            if_empty=self.if_empty,
            flow=self,
        )

        table.set_upstream(parquet, flow=self)

        if self.update_kv == True:
            set_new_kv.bind(
                kv_name=self.name,
                df=df,
                filter_column=self.filter_column,
                flow=self,
            )
            set_new_kv.set_upstream(table, flow=self)

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()



import os
from pathlib import Path
from typing import Any, Dict, List, Literal

import pandas as pd
import pendulum
from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from viadot.tasks import AzureDataLakeUpload, VidClubToDF

logger = logging.get_logger(__name__)


class VidClubToADLS(Flow):
    def __init__(
        self,
        name: str,
        source: Literal["jobs", "product", "company", "survey"] = None,
        from_date: str = "2022-03-22",
        to_date: str = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] = "all",
        days_interval: int = 30,
        cols_to_drop: List[str] = None,
        vid_club_credentials: Dict[str, Any] = None,
        vidclub_credentials_secret: str = "VIDCLUB",
        vidclub_vault_name: str = None,
        output_file_extension: str = ".parquet",
        adls_dir_path: str = None,
        local_file_path: str = None,
        adls_file_name: str = None,
        vault_name: str = None,
        adls_sp_credentials_secret: str = None,
        overwrite_adls: bool = False,
        if_exists: str = "replace",
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any]
    ):
        """
        Flow for downloading data from the Vid Club via API to a CSV or Parquet file.
        Then upload it to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None, which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page. Defaults to 100.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional): Region filter for the query. Defaults to "all". [July 2023 status: parameter works only for 'all' on API]
            days_interval (int, optional): Days specified in date range per API call (test showed that 30-40 is optimal for performance). Defaults to 30.
            cols_to_drop (List[str], optional): List of columns to drop. Defaults to None.
            vid_club_credentials (Dict[str, Any], optional): Stores the credentials information. Defaults to None.
            vidclub_credentials_secret (str, optional): The name of the secret in Azure Key Vault or Prefect or local_config file. Defaults to "VIDCLUB".
            vidclub_vault_name (str, optional): For Vid Club credentials stored in Azure Key Vault. The name of the vault from which to obtain the secret. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data
                which is not easy to handle with parquet. Defaults to ".parquet".
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            vault_name (str, optional): For ADLS credentials stored in Azure Key Vault. The name of the vault from which to obtain the secret. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
                Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
            timeout (int, optional): The time (in seconds) to wait while running this task before a timeout occurs. Defaults to 3600.
        """
        # VidClubToDF
        self.source = source
        self.from_date = from_date
        self.to_date = to_date
        self.items_per_page = items_per_page
        self.region = region
        self.days_interval = days_interval
        self.cols_to_drop = cols_to_drop
        self.vid_club_credentials = vid_club_credentials
        self.vidclub_credentials_secret = vidclub_credentials_secret
        self.vidclub_vault_name = vidclub_vault_name

        # AzureDataLakeUpload
        self.adls_file_name = adls_file_name
        self.adls_dir_path = adls_dir_path
        self.local_file_path = local_file_path
        self.overwrite = overwrite_adls
        self.vault_name = vault_name
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.timeout = timeout
        self.now = str(pendulum.now("utc"))

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        vid_club_df_task = VidClubToDF(
            timeout=self.timeout,
            source=self.source,
            credentials=self.vid_club_credentials,
            credentials_secret=self.vidclub_credentials_secret,
            vault_name=self.vidclub_vault_name,
        )

        vid_club_df = vid_club_df_task.bind(
            from_date=self.from_date,
            to_date=self.to_date,
            items_per_page=self.items_per_page,
            region=self.region,
            days_interval=self.days_interval,
            cols_to_drop=self.cols_to_drop,
            flow=self,
        )

        df_with_metadata = add_ingestion_metadata_task.bind(vid_club_df, flow=self)

        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)
        df_mapped = df_map_mixed_dtypes_for_parquet.bind(
            df_with_metadata, dtypes_dict, flow=self
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_mapped,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )
        json_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



from unittest import mock

import pytest

from viadot.sources import Genesys


@pytest.fixture
def var_dictionary():
    variables = {
        "start_date": "2022-08-12",
        "media_type_list": ["callback", "chat"],
        "queueIds_list": [
            "1234567890",
            "1234567890",
        ],
        "data_to_post": """{
                "name": f"QUEUE_PERFORMANCE_DETAIL_VIEW_{media}",
                "timeZone": "UTC",
                "exportFormat": "CSV",
                "interval": f"{end_date}T23:00:00/{start_date}T23:00:00",
                "period": "PT30M",
                "viewType": f"QUEUE_PERFORMANCE_DETAIL_VIEW",
                "filter": {"mediaTypes": [f"{media}"], "queueIds": [f"{queueid}"], "directions":["inbound"],},
                "read": True,
                "locale": "en-us",
                "hasFormatDurations": False,
                "hasSplitFilters": True,
                "excludeEmptyRows": True,
                "hasSplitByMedia": True,
                "hasSummaryRow": True,
                "csvDelimiter": "COMMA",
                "hasCustomParticipantAttributes": True,
                "recipientEmails": [],
            }""",
        "report_data": [
            [
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "https://apps.mypurecloud.de/example/url/test",
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "chat",
                "QUEUE_PERFORMANCE_DETAIL_VIEW",
                "2022-08-12T23:00:00.000Z/2022-08-13T23:00:00.000Z",
                "COMPLETED",
            ],
            [
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "https://apps.mypurecloud.de/example/url/test",
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "chat",
                "QUEUE_PERFORMANCE_DETAIL_VIEW",
                "2022-08-12T23:00:00.000Z/2022-08-13T23:00:00.000Z",
                "COMPLETED",
            ],
            [
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "https://apps.mypurecloud.de/example/url/test",
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "callback",
                "QUEUE_PERFORMANCE_DETAIL_VIEW",
                "2022-08-12T23:00:00.000Z/2022-08-13T23:00:00.000Z",
                "COMPLETED",
            ],
            [
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "https://apps.mypurecloud.de/example/url/test",
                "1234567890qwertyuiopasdfghjklazxcvbn",
                "callback",
                "QUEUE_PERFORMANCE_DETAIL_VIEW",
                "2022-08-12T23:00:00.000Z/2022-08-13T23:00:00.000Z",
                "COMPLETED",
            ],
        ],
        "entities": {
            "entities": [
                {
                    "id": "1234567890",
                    "name": "QUEUE_PERFORMANCE_DETAIL_VIEW_chat",
                    "runId": "1234567890",
                    "status": "COMPLETED",
                    "timeZone": "UTC",
                    "exportFormat": "CSV",
                    "interval": "2022-08-02T23:00:00.000Z/2022-08-03T23:00:00.000Z",
                    "downloadUrl": "https://apps.mypurecloud.de/example/url/test",
                    "viewType": "QUEUE_PERFORMANCE_DETAIL_VIEW",
                    "period": "PT30M",
                    "filter": {
                        "mediaTypes": ["chat"],
                        "queueIds": ["1234567890"],
                        "directions": ["inbound"],
                    },
                    "read": False,
                    "createdDateTime": "2022-08-03T11:19:47Z",
                    "modifiedDateTime": "2022-08-03T11:19:49Z",
                    "locale": "en-us",
                    "percentageComplete": 1.0,
                    "hasFormatDurations": False,
                    "hasSplitFilters": True,
                    "excludeEmptyRows": True,
                    "hasSplitByMedia": True,
                    "hasSummaryRow": True,
                    "csvDelimiter": "COMMA",
                    "hasCustomParticipantAttributes": True,
                    "dateLastConfirmed": "2022-08-03T11:19:47Z",
                    "intervalKeyType": "ConversationStart",
                    "enabled": False,
                    "selfUri": "/api/v2/example/url/test",
                },
            ],
            "pageSize": 100,
            "pageNumber": 1,
            "total": 6,
            "pageCount": 1,
        },
        "ids_mapping": {"1234567890qwertyuiopasdfghjklazxcvbn": "TEST"},
    }

    return variables


class MockClass:
    status_code = 200

    def json():
        test = {"token_type": None, "access_token": None}
        return test


@pytest.mark.init
def test_create_genesys_class():
    g = Genesys()
    assert g


@pytest.mark.init
def test_default_credential_param():
    g = Genesys()
    assert g.credentials != None and type(g.credentials) == dict


@pytest.mark.init
def test_environment_param():
    g = Genesys()
    assert g.environment != None and type(g.environment) == str


@pytest.mark.parametrize("input_name", ["test_name", "12345", ".##@@"])
@pytest.mark.init
def test_other_inicial_params(input_name):
    g = Genesys(report_name=input_name)
    assert len(g.report_name) > 0 and isinstance(g.report_name, str)


@pytest.mark.proper
def test_connection_with_genesys_api():
    g = Genesys()
    test_genesys_connection = g.authorization_token
    assert (
        type(test_genesys_connection) == dict
        and len(test_genesys_connection.items()) > 0
    )


@mock.patch.object(Genesys, "genesys_api_connection")
@pytest.mark.connection
def test_generate_api_connection(mock_api_response, var_dictionary):
    g = Genesys()
    assert g.genesys_api_connection()
    mock_api_response.assert_called()


@mock.patch.object(Genesys, "download_report")
@pytest.mark.dependency(depends=["test_generate_api_connection"])
@pytest.mark.download
def test_download_reports(mock_download_files, var_dictionary):
    g = Genesys()
    g.ids_mapping = var_dictionary["ids_mapping"]
    g.report_data = var_dictionary["report_data"]
    g.start_date = var_dictionary["start_date"]
    file_name_list = g.download_all_reporting_exports()

    assert type(file_name_list) == list and len(file_name_list) > 0
    mock_download_files.assert_called()


@mock.patch("viadot.sources.genesys.handle_api_response", return_value=MockClass)
@pytest.mark.dependency(
    depends=[
        "test_generate_api_connection",
        "test_download_reports",
    ]
)
@pytest.mark.delete
def test_genesys_delete_reports(mock_api_response, var_dictionary):
    g = Genesys()
    g.report_data = var_dictionary["report_data"]
    g.delete_all_reporting_exports()
    mock_api_response.assert_called()



from unittest import mock

import pytest
from prefect.engine.state import State, Success

from viadot.flows import Pipeline


def test_pipeline_init():
    instance = Pipeline(
        "test_pipeline_flow",
        project_name="example_project",
        extract_flows_names=["flow1_extract", "flow2_load"],
        transform_flow_name="flow1_extract",
    )
    assert instance


def test_pipeline_flow_run_mock():
    with mock.patch.object(Pipeline, "run", return_value=Success) as mock_method:
        instance = Pipeline(
            "test_pipeline_flow",
            project_name="example_project",
            extract_flows_names=["flow1_extract", "flow2_load"],
            transform_flow_name="flow1_extract",
        )
        result = instance.run()
        assert result



import pytest

from viadot.flows.sql_server_transform import SQLServerTransform
from viadot.tasks.sql_server import SQLServerQuery

SCHEMA = "sandbox"
TABLE = "test_sql_server"


def test_sql_server_transform():
    query_task = SQLServerQuery("AZURE_SQL")
    query_task.run(f"DROP TABLE IF EXISTS {SCHEMA}.{TABLE}")
    flow = SQLServerTransform(
        name="test flow sql transform",
        config_key="AZURE_SQL",
        query=f"CREATE TABLE {SCHEMA}.{TABLE} (Id INT, Name VARCHAR (10))",
    )
    result = flow.run()
    assert result.is_successful()
    query_task.run(f"DROP TABLE {SCHEMA}.{TABLE}")


def test_sql_server_transfor_fail():
    flow = SQLServerTransform(
        name="test flow sql transform",
        config_key="AZURE_SQL",
        query=f"CREATE TABLE {SCHEMA}.{TABLE}",
    )

    result = flow.run()
    assert result.is_failed()



from unittest import mock

import pandas as pd
import pytest

from viadot.config import local_config
from viadot.sources import Outlook
from viadot.tasks import OutlookToDF

outlook_env_vars = local_config.get("OUTLOOK")


def to_df():
    test_df = pd.DataFrame(
        data={"country": ["italy", "germany", "spain"], "sales": [100, 50, 80]}
    )

    assert isinstance(test_df, pd.DataFrame)

    return test_df


@pytest.mark.init
def test_outlook_to_df():
    outlook = Outlook(
        mailbox_name=outlook_env_vars["mail_example"],
        credentials=outlook_env_vars,
        start_date="2022-04-28",
        end_date="2022-04-29",
    )
    df = outlook.get_all_mails_to_df()

    assert isinstance(df, pd.DataFrame)


@pytest.mark.init
@pytest.mark.dependency(depends=["test_outlook_to_df"])
def test_outlook_credentials():
    outlook_env_vars = local_config.get("OUTLOOK")
    outlook = Outlook(
        mailbox_name=outlook_env_vars["mail_example"], credentials=outlook_env_vars
    )

    assert isinstance(outlook.credentials, dict)


@mock.patch("O365.Account", return_value=None)
@mock.patch("O365.Account.authenticate", return_value="trial1")
@mock.patch("O365.Account.mailbox", return_value="trial2")
@mock.patch("viadot.sources.Outlook.get_all_mails_to_df", return_calue=to_df)
@pytest.mark.task
def test_outlook_task(mock_account, mock_auth, mock_mail, mock_to_df):
    outlook_to_df = OutlookToDF(credentials=outlook_env_vars)
    df = outlook_to_df.run(
        mailbox_name=outlook_env_vars["mail_example"],
    )

    assert df != None
    mock_account.assert_called()
    mock_auth.assert_called()
    mock_mail.assert_called()



from unittest import mock

from viadot.flows import ADLSGen1ToAzureSQL


def test_adls_gen1_to_azure_sql_new_init(
    TEST_CSV_FILE_BLOB_PATH, TEST_PARQUET_FILE_PATH
):
    instance = ADLSGen1ToAzureSQL(
        name="test_adls_gen1_azure_sql_flow",
        path=TEST_PARQUET_FILE_PATH,
        blob_path=TEST_CSV_FILE_BLOB_PATH,
        schema="sandbox",
        table="test_bcp",
        dtypes={"country": "VARCHAR(25)", "sales": "INT"},
        if_exists="replace",
    )
    assert instance


def test_adls_gen1_to_azure_sql_new_mock(
    TEST_CSV_FILE_BLOB_PATH, TEST_PARQUET_FILE_PATH
):
    with mock.patch.object(ADLSGen1ToAzureSQL, "run", return_value=True) as mock_method:
        instance = ADLSGen1ToAzureSQL(
            name="test_adls_gen1_azure_sql_flow",
            path=TEST_PARQUET_FILE_PATH,
            blob_path=TEST_CSV_FILE_BLOB_PATH,
            schema="sandbox",
            table="test_bcp",
            dtypes={"country": "VARCHAR(25)", "sales": "INT"},
            if_exists="replace",
        )
        instance.run()
        mock_method.assert_called_with()



import logging
import os

import pytest
from prefect.storage import Local

from viadot.flows import SupermetricsToADLS

CWD = os.getcwd()
adls_dir_path = "raw/supermetrics"
STORAGE = Local(path=CWD)

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def expectation_suite():
    expectation_suite = {
        "data_asset_type": "Dataset",
        "expectation_suite_name": "failure",
        "expectations": [
            {
                "expectation_type": "expect_table_row_count_to_be_between",
                "kwargs": {
                    "max_value": {"$PARAMETER": "trunc(previous_run_row_count * 1.2)"},
                    "min_value": {"$PARAMETER": "trunc(previous_run_row_count * 0.8)"},
                },
                "meta": {},
            },
        ],
        "meta": {
            "columns": {
                "All Users": {"description": ""},
                "Date": {"description": ""},
                "M-Site_Better Space: All Landing Page Sessions": {"description": ""},
                "M-site_Accessories: All Landing Page Sessions": {"description": ""},
                "M-site_More Space: All Landing Page Sessions": {"description": ""},
                "M-site_Replacement: All Landing Page Sessions": {"description": ""},
            },
            "great_expectations_version": "0.13.19",
        },
    }

    yield expectation_suite


def test_supermetrics_to_adls(expectation_suite):
    flow = SupermetricsToADLS(
        "Google Analytics Load Times extract test",
        ds_id="GA",
        ds_segments=[
            "R1fbzFNQQ3q_GYvdpRr42w",
            "I8lnFFvdSFKc50lP7mBKNA",
            "Lg7jR0VWS5OqGPARtGYKrw",
            "h8ViuGLfRX-cCL4XKk6yfQ",
            "-1",
        ],
        ds_accounts=["8326007", "58338899"],
        date_range_type="last_year_inc",
        fields=[
            {"id": "Date"},
            {"id": "segment", "split": "column"},
            {"id": "AvgPageLoadTime_calc"},
        ],
        settings={"avoid_sampling": "true"},
        order_columns="alphabetic",
        max_columns=100,
        max_rows=10,
        expectation_suite=expectation_suite,
        evaluation_parameters=dict(previous_run_row_count=9),
        adls_dir_path=adls_dir_path,
        parallel=False,
        storage=STORAGE,
    )
    result = flow.run()
    assert result.is_successful()

    task_results = result.result.values()
    assert all([task_result.is_successful() for task_result in task_results])


def test_supermetrics_to_adls_file_name(expectation_suite):
    flow = SupermetricsToADLS(
        "Google Analytics Load Times extract test file name",
        ds_id="GA",
        ds_segments=[
            "R1fbzFNQQ3q_GYvdpRr42w",
            "I8lnFFvdSFKc50lP7mBKNA",
            "Lg7jR0VWS5OqGPARtGYKrw",
            "h8ViuGLfRX-cCL4XKk6yfQ",
            "-1",
        ],
        ds_accounts=["8326007", "58338899"],
        date_range_type="last_year_inc",
        fields=[
            {"id": "Date"},
            {"id": "segment", "split": "column"},
            {"id": "AvgPageLoadTime_calc"},
        ],
        settings={"avoid_sampling": "true"},
        order_columns="alphabetic",
        max_columns=100,
        max_rows=10,
        evaluation_parameters=dict(previous_run_row_count=9),
        adls_dir_path=adls_dir_path,
        expectation_suite=expectation_suite,
        adls_file_name="test_file.csv",
        parallel=False,
        storage=STORAGE,
    )
    result = flow.run()
    assert result.is_successful()



import pandas as pd

from viadot.tasks import ASELiteToDF


def test_aselite_to_df():
    query = """SELECT TOP (10) [usageid]
      ,[configid]
      ,[verticalid]
      ,[textgroupid]
      ,[nr]
      ,[storedate]
  FROM [UCRMDEV_DESIGNER].[dbo].[PORTAL_APPLICATION_TEXTUSAGE]"""
    task = ASELiteToDF()
    df = task.run(query=query)
    assert isinstance(df, pd.DataFrame)



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import DuckDBTransform
from viadot.sources import DuckDB
from viadot.tasks import DuckDBQuery, DuckDBToDF

TABLE = "test_table"
SCHEMA = "test_schema"
TABLE_MULTIPLE_PARQUETS = "test_multiple_parquets"
DATABASE_PATH = "test_db_123.duckdb"


@pytest.fixture(scope="session")
def duckdb():
    duckdb = DuckDB(credentials=dict(database=DATABASE_PATH))
    yield duckdb
    os.remove(DATABASE_PATH)


def test_create_table_from_parquet(duckdb, TEST_PARQUET_FILE_PATH):
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
    )


def test_duckdb_query():
    db_query = DuckDBQuery(credentials=dict(database=DATABASE_PATH))

    result = db_query.run(f"select * from {SCHEMA}.{TABLE}")
    assert type(result) == list
    assert len(result) > 1


def test_duckdb_to_df():
    instance = DuckDBToDF(
        schema=SCHEMA, table=TABLE, credentials=dict(database=DATABASE_PATH)
    )
    test_df = instance.run()
    assert test_df.shape > (1, 1)
    assert type(test_df) == pd.core.frame.DataFrame


def test_duckdb_transform_init():
    instance = DuckDBTransform("test_duckdb_transform", query="select * from test")

    assert instance


def test_duckdb_transform_flow_run():
    instance = DuckDBTransform(
        "test_duckdb_transform",
        query=f"select * from {SCHEMA}.{TABLE}",
        credentials=dict(database=DATABASE_PATH),
    )
    result = instance.run()
    assert result.is_successful()



import logging

import pandas as pd
import pytest

from viadot.exceptions import ValidationError
from viadot.tasks import AzureSQLCreateTable, AzureSQLDBQuery, CheckColumnOrder

logger = logging.getLogger(__name__)

SCHEMA = "sandbox"
TABLE = "test"


def test_azure_sql_create_table():
    create_table_task = AzureSQLCreateTable()

    create_table_task.run(
        schema=SCHEMA,
        table=TABLE,
        dtypes={"id": "INT", "name": "VARCHAR(25)"},
        if_exists="replace",
    )


def test_azure_sql_run_sqldb_query_empty_result():
    sql_query_task = AzureSQLDBQuery()

    list_table_info_query = f"""
        SELECT *
        FROM sys.tables t
        JOIN sys.schemas s
            ON t.schema_id = s.schema_id
        WHERE s.name = '{SCHEMA}' AND t.name = '{TABLE}'
    """
    exists = bool(sql_query_task.run(list_table_info_query))
    assert exists

    result = sql_query_task.run(f"SELECT * FROM {SCHEMA}.{TABLE}")
    assert result == []


def test_azure_sql_run_insert_query():
    sql_query_task = AzureSQLDBQuery()

    sql_query_task.run(f"INSERT INTO {SCHEMA}.{TABLE} VALUES (1, 'Mike')")
    result = list(sql_query_task.run(f"SELECT * FROM {SCHEMA}.{TABLE}")[0])
    assert result == [1, "Mike"]


def test_azure_sql_run_drop_query():
    sql_query_task = AzureSQLDBQuery()

    result = sql_query_task.run(f"DROP TABLE {SCHEMA}.{TABLE}")
    assert result is True

    list_table_info_query = f"""
        SELECT *
        FROM sys.tables t
        JOIN sys.schemas s
            ON t.schema_id = s.schema_id
        WHERE s.name = '{SCHEMA}' AND t.name = '{TABLE}'
    """
    exists = bool(sql_query_task.run(list_table_info_query))
    assert not exists


def test_check_column_order_append_same_col_number(caplog):
    create_table_task = AzureSQLCreateTable()
    with caplog.at_level(logging.INFO):
        create_table_task.run(
            schema=SCHEMA,
            table=TABLE,
            dtypes={"id": "INT", "name": "VARCHAR(25)", "street": "VARCHAR(25)"},
            if_exists="replace",
        )
    assert "Successfully created table sandbox" in caplog.text

    data = {"id": [1], "street": ["Green"], "name": ["Tom"]}
    df = pd.DataFrame(data)

    check_column_order = CheckColumnOrder()
    with caplog.at_level(logging.WARNING):
        check_column_order.run(table=TABLE, schema=SCHEMA, if_exists="append", df=df)

    assert (
        "Detected column order difference between the CSV file and the table. Reordering..."
        in caplog.text
    )


def test_check_column_order_append_diff_col_number(caplog):
    create_table_task = AzureSQLCreateTable()
    with caplog.at_level(logging.INFO):
        create_table_task.run(
            schema=SCHEMA,
            table=TABLE,
            dtypes={"id": "INT", "name": "VARCHAR(25)", "street": "VARCHAR(25)"},
            if_exists="replace",
        )
    assert "Successfully created table sandbox" in caplog.text

    data = {"id": [1], "age": ["40"], "street": ["Green"], "name": ["Tom"]}
    df = pd.DataFrame(data)
    print(f"COMP: \ndf: {df.columns} \nsql: ")
    check_column_order = CheckColumnOrder()
    with pytest.raises(
        ValidationError,
        match=r"Detected discrepancies in number of columns or different column names between the CSV file and the SQL table!",
    ):
        check_column_order.run(table=TABLE, schema=SCHEMA, if_exists="append", df=df)


def test_check_column_order_replace(caplog):
    create_table_task = AzureSQLCreateTable()
    with caplog.at_level(logging.INFO):
        create_table_task.run(
            schema=SCHEMA,
            table=TABLE,
            dtypes={"id": "INT", "name": "VARCHAR(25)", "street": "VARCHAR(25)"},
            if_exists="replace",
        )
    assert "Successfully created table sandbox" in caplog.text

    data = {"id": [1], "street": ["Green"], "name": ["Tom"]}
    df = pd.DataFrame(data)

    check_column_order = CheckColumnOrder()
    with caplog.at_level(logging.INFO):
        check_column_order.run(table=TABLE, schema=SCHEMA, if_exists="replace", df=df)
    assert "The table will be replaced." in caplog.text


def test_check_column_order_append_not_exists(caplog):
    check_column_order = CheckColumnOrder()
    data = {"id": [1], "street": ["Green"], "name": ["Tom"]}
    df = pd.DataFrame(data)
    check_column_order.run(
        table="non_existing_table_123", schema="sandbox", if_exists="append", df=df
    )
    assert "table doesn't exists" in caplog.text



from prefect import Task

from ..sources import AzureBlobStorage

# from ..tasks import AzureKeyVaultSecret


class BlobFromCSV(Task):
    """
    Task for generating Azure Blob Storage from CSV file
    """

    def __init__(self, timeout: int = 3600, *args, **kwargs):
        super().__init__(name="csv_to_blob_storage", timeout=timeout, *args, **kwargs)

    def __call__(self):
        """Generate a blob from a local CSV file"""

    def run(
        self,
        from_path: str,
        to_path: str,
        overwrite: bool = False,
        sp_credentials_secret: str = None,
        vault_name: str = None,
    ):
        """[summary]

        Args:
            from_path (str): The path from which to download the file(s). Defaults to None.
            to_path (str): The destination path. Defaults to None.
            overwrite (bool): Indicator what to do if file already exists. Defaults to False.
            sp_credentials_secret (str, optional): The name of the Azure KeyVault secret containing a dictionary with
            Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        # if sp_credentials_secret:
        #     azure_secret_task = AzureKeyVaultSecret()
        #     credentials = azure_secret_task.run(
        #         secret=sp_credentials_secret, vault_name=vault_name
        #     )
        # else:
        #     credentials = None
        # blob_storage = AzureBlobStorage(credentials=credentials)

        blob_storage = AzureBlobStorage()

        self.logger.info(f"Copying from {from_path} to {to_path}...")

        result = blob_storage.to_storage(
            from_path=from_path, to_path=to_path, overwrite=overwrite
        )
        self.logger.info(f"Successfully uploaded data to {to_path}.")

        return result



import xml.etree.ElementTree as ET
from typing import Any, Dict, Optional

import pandas as pd
from pydantic import BaseModel

from ..config import local_config
from ..exceptions import CredentialError, DataRangeError
from ..utils import handle_api_response
from .base import Source

""" 
The official documentation does not specify the list of required 
fields so they were set as optional in BaseModel classes.
"""


class TrackingNumbers(BaseModel):
    TrackingNumber: Optional[str]


class ShipToAddress(BaseModel):
    ShipToNumber: Optional[str]
    Attention: Optional[str]
    AddressLine1: Optional[str]
    AddressLine2: Optional[str]
    AddressLine3: Optional[str]
    City: Optional[str]
    State: Optional[str]
    Zip: Optional[str]
    Country: Optional[str]
    EmailAddress: Optional[str]
    PhoneNumber: Optional[str]
    FaxNumber: Optional[str]


class InvoiceTotals(BaseModel):
    Merchandise: Optional[str]
    InboundFreight: Optional[str]
    OutboundFreight: Optional[str]
    Handling: Optional[str]
    Delivery: Optional[str]
    Pickup: Optional[str]
    Restocking: Optional[str]
    MinimumCharge: Optional[str]
    DiscountAllowance: Optional[str]
    SalesTax: Optional[str]
    TotalInvoice: Optional[str]


class HeaderInformation(BaseModel):
    CompanyNumber: Optional[str]
    OrderNumber: Optional[str]
    InvoiceNumber: Optional[str]
    CustomerNumber: Optional[str]
    CustomerDescription: Optional[str]
    CustomerPurchaseOrderNumber: Optional[str]
    Contact: Optional[str]
    SellingWarehouse: Optional[str]
    ShippingWarehouse: Optional[str]
    ShippingMethod: Optional[str]
    PaymentTerms: Optional[str]
    PaymentTermsDescription: Optional[str]
    FreightTerms: Optional[str]
    FreightTermsDescription: Optional[str]
    SalesRepOne: Optional[str]
    SalesRepOneDescription: Optional[str]
    EntryDate: Optional[str]
    OrderDate: Optional[str]
    RequiredDate: Optional[str]
    ShippedDate: Optional[str]
    InvoiceDate: Optional[str]
    ShipToAddress: Optional[ShipToAddress]
    TrackingNumbers: Optional[TrackingNumbers]
    InvoiceTotals: Optional[InvoiceTotals]


class LineItemDetail(BaseModel):
    ProductNumber: Optional[str]
    ProductDescription1: Optional[str]
    ProductDescription2: Optional[str]
    CustomerProductNumber: Optional[str]
    LineItemNumber: Optional[str]
    QuantityOrdered: Optional[str]
    QuantityShipped: Optional[str]
    QuantityBackordered: Optional[str]
    Price: Optional[str]
    UnitOfMeasure: Optional[str]
    ExtendedPrice: Optional[str]
    QuantityShippedExtension: Optional[str]
    LineItemShipWarehouse: Optional[str]
    RequiredDate: Optional[str]
    CopperWeight: Optional[str]


class Order(BaseModel):
    HeaderInformation: Optional[HeaderInformation]
    LineItemDetail: Optional[LineItemDetail]


def parse_orders_xml(xml_data: str) -> pd.DataFrame:
    """
    Function to parse xml containing Epicor Orders Data.

    Args:
        xml_data (str, required): Response from Epicor API in form of xml
    Returns:
        pd.DataFrame: DataFrame containing parsed orders data.
    """
    final_df = pd.DataFrame()
    ship_dict = {}
    invoice_dict = {}
    header_params_dict = {}
    item_params_dict = {}

    root = ET.fromstring(xml_data.text)

    for order in root.findall("Order"):
        for header in order.findall("HeaderInformation"):
            for tracking_numbers in header.findall("TrackingNumbers"):
                numbers = ""
                for tracking_number in tracking_numbers.findall("TrackingNumber"):
                    numbers = numbers + "'" + tracking_number.text + "'"
                result_numbers = TrackingNumbers(TrackingNumber=numbers)

            for shipto in header.findall("ShipToAddress"):
                for ship_param in ShipToAddress.__dict__.get("__annotations__"):
                    try:
                        ship_value = shipto.find(ship_param).text
                    except:
                        ship_value = None
                    ship_parameter = {ship_param: ship_value}
                    ship_dict.update(ship_parameter)
                ship_address = ShipToAddress(**ship_dict)

            for invoice in header.findall("InvoiceTotals"):
                for invoice_param in InvoiceTotals.__dict__.get("__annotations__"):
                    try:
                        invoice_value = invoice.find(invoice_param).text
                    except:
                        invoice_value = None
                    invoice_parameter = {invoice_param: invoice_value}
                    invoice_dict.update(invoice_parameter)
                invoice_total = InvoiceTotals(**invoice_dict)

            for header_param in HeaderInformation.__dict__.get("__annotations__"):
                try:
                    header_value = header.find(header_param).text
                except:
                    header_value = None
                if header_param == "TrackingNumbers":
                    header_parameter = {header_param: result_numbers}
                elif header_param == "ShipToAddress":
                    header_parameter = {header_param: ship_address}
                elif header_param == "InvoiceTotals":
                    header_parameter = {header_param: invoice_total}
                else:
                    header_parameter = {header_param: header_value}
                header_params_dict.update(header_parameter)
            header_info = HeaderInformation(**header_params_dict)
        for items in order.findall("LineItemDetails"):
            for item in items.findall("LineItemDetail"):
                for item_param in LineItemDetail.__dict__.get("__annotations__"):
                    try:
                        item_value = item.find(item_param).text
                    except:
                        item_value = None
                    item_parameter = {item_param: item_value}
                    item_params_dict.update(item_parameter)
                line_item = LineItemDetail(**item_params_dict)
                row = Order(HeaderInformation=header_info, LineItemDetail=line_item)
                my_dict = row.dict()
                final_df = final_df.append(
                    pd.json_normalize(my_dict, max_level=2), ignore_index=True
                )
    return final_df


class Epicor(Source):
    def __init__(
        self,
        base_url: str,
        filters_xml: str,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        validate_date_filter: bool = True,
        start_date_field: str = "BegInvoiceDate",
        end_date_field: str = "EndInvoiceDate",
        *args,
        **kwargs,
    ):
        """
        Class to connect to Epicor API and download results.

        Args:
            base_url (str, required): Base url to Epicor Orders.
            filters_xml (str, required): Filters in form of XML. The date filter is required.
            credentials (Dict[str, Any], optional): Credentials to connect with Epicor API containing host, port, username and password.
                Defaults to None.
            config_key (str, optional): Credential key to dictionary where details are stored.
            validate_date_filter (bool, optional): Whether or not validate xml date filters. Defaults to True.
            start_date_field (str, optional) The name of filters field containing start date. Defaults to "BegInvoiceDate".
            end_date_field (str, optional) The name of filters field containing end date. Defaults to "EndInvoiceDate".
        """
        DEFAULT_CREDENTIALS = local_config.get(config_key)
        credentials = credentials or DEFAULT_CREDENTIALS

        required_credentials = ["host", "port", "username", "password"]
        if any([cred_key not in credentials for cred_key in required_credentials]):
            not_found = [c for c in required_credentials if c not in credentials]
            raise CredentialError(f"Missing credential(s): '{not_found}'.")

        self.credentials = credentials
        self.config_key = config_key
        self.base_url = base_url
        self.filters_xml = filters_xml
        self.validate_date_filter = validate_date_filter
        self.start_date_field = start_date_field
        self.end_date_field = end_date_field

        super().__init__(*args, credentials=credentials, **kwargs)

    def generate_token(self) -> str:
        "Function to generate API access token that is valid for 24 hours"

        url = (
            "http://"
            + self.credentials["host"]
            + ":"
            + str(self.credentials["port"])
            + "/api/security/token/"
        )

        headers = {
            "Content-Type": "application/json",
            "username": self.credentials["username"],
            "password": self.credentials["password"],
        }

        response = handle_api_response(url=url, headers=headers, method="POST")
        root = ET.fromstring(response.text)
        token = root.find("AccessToken").text
        return token

    def generate_url(self) -> str:
        "Function to generate url to download data"

        return (
            "http://"
            + self.credentials["host"]
            + ":"
            + str(self.credentials["port"])
            + self.base_url
        )

    def validate_filter(self) -> None:
        "Function checking if user had specified date range filters."

        root = ET.fromstring(self.filters_xml)
        for child in root:
            for subchild in child:
                if (
                    subchild.tag == self.start_date_field
                    or subchild.tag == self.end_date_field
                ) and subchild.text == None:
                    raise DataRangeError(
                        "Too much data. Please provide a date range filter."
                    )

    def get_xml_response(self):
        "Function for getting response from Epicor API"
        if self.validate_date_filter == True:
            self.validate_filter()
        payload = self.filters_xml
        url = self.generate_url()
        headers = {
            "Content-Type": "application/xml",
            "Authorization": "Bearer " + self.generate_token(),
        }
        response = handle_api_response(
            url=url, headers=headers, body=payload, method="POST"
        )
        return response

    def to_df(self):
        "Function for creating pandas DataFrame from Epicor API response"
        data = self.get_xml_response()
        df = parse_orders_xml(data)
        return df



import os

import pytest

from viadot.config import local_config
from viadot.sources import AzureBlobStorage, AzureSQL, Supermetrics

SUPERMETRICS_CREDENTIALS = local_config.get("SUPERMETRICS")
FILE_PATH = "/home/viadot/tests/integration/test_supermetrics.csv"
BLOB_PATH = "tests/supermetrics/test_supermetrics.csv"

SCHEMA = "sandbox"
TABLE = "test_supermetrics"


@pytest.fixture
def supermetrics():
    supermetrics = Supermetrics(config_key="SUPERMETRICS")
    yield supermetrics


@pytest.fixture
def azstorage():
    azstorage = AzureBlobStorage(config_key="AZURE_BLOB_STORAGE")
    yield azstorage
    os.remove(FILE_PATH)


# Replace these hacks later
# Date
dtypes = {
    "Date": "DATE",
    "profile": "VARCHAR(255)",
    "Campaignname": "VARCHAR(255)",
    "Impressions": "FLOAT(24)",
    "Clicks": "FLOAT(24)",
    "Cost_eur": "FLOAT(24)",
    "SearchImpressionShare": "VARCHAR(255)",
}
# End crappy hacks

sm_query = {
    "ds_id": "AW",
    "ds_accounts": SUPERMETRICS_CREDENTIALS["SOURCES"]["Google Ads"]["Accounts"],
    "ds_user": SUPERMETRICS_CREDENTIALS["USER"],
    "date_range_type": "last_year_inc",
    "fields": [
        "Date",
        "profile",
        "Campaignname",
        "Impressions",
        "Clicks",
        "Cost_eur",
        "SearchImpressionShare",
    ],
    "max_rows": 1000000,
}


def test_fetch_supermetrics(supermetrics):
    supermetrics.query(sm_query)
    supermetrics.to_csv(FILE_PATH)


def test_upload(azstorage):
    azstorage.to_storage(from_path=FILE_PATH, to_path=BLOB_PATH, overwrite=True)


def test_bulk_insert():
    azuresql = AzureSQL(config_key="AZURE_SQL")

    azuresql.create_table(
        schema=SCHEMA, table=TABLE, dtypes=dtypes, if_exists="replace"
    )
    azuresql.bulk_insert(
        table=TABLE,
        schema=SCHEMA,
        source_path=BLOB_PATH,
        if_exists="replace",
    )



import json
import os
import urllib
from datetime import date, datetime, timedelta
from typing import Any, Dict, List, Literal, Tuple

import pandas as pd
from prefect.utilities import logging

from ..exceptions import CredentialError, ValidationError
from ..utils import handle_api_response
from .base import Source

logger = logging.get_logger()


class VidClub(Source):
    """
    A class implementing the Vid Club API.

    Documentation for this API is located at: https://evps01.envoo.net/vipapi/
    There are 4 endpoints where to get the data.
    """

    def __init__(self, credentials: Dict[str, Any], *args, **kwargs):
        """
        Create an instance of VidClub.

        Args:
            credentials (Dict[str, Any]): Credentials to Vid Club APIs containing token.

        Raises:
            CredentialError: If credentials are not provided as a parameter.
        """
        self.headers = {
            "Authorization": "Bearer " + credentials["token"],
            "Content-Type": "application/json",
        }

        super().__init__(*args, credentials=credentials, **kwargs)

    def build_query(
        self,
        from_date: str,
        to_date: str,
        api_url: str,
        items_per_page: int,
        source: Literal["jobs", "product", "company", "survey"] = None,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] = "all",
    ) -> str:
        """
        Builds the query from the inputs.

        Args:
            from_date (str): Start date for the query.
            to_date (str): End date for the query, if empty, will be executed as datetime.today().strftime("%Y-%m-%d").
            api_url (str): Generic part of the URL to Vid Club API.
            items_per_page (int): number of entries per page.
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional): Region filter for the query. Defaults to "all". [July 2023 status: parameter works only for 'all' on API]

        Returns:
            str: Final query with all filters added.

        Raises:
            ValidationError: If any source different than the ones in the list are used.
        """
        if source in ["jobs", "product", "company"]:
            url = f"{api_url}{source}?from={from_date}&to={to_date}&region={region}&limit={items_per_page}"
        elif source == "survey":
            url = f"{api_url}{source}?language=en&type=question"
        else:
            raise ValidationError(
                "Pick one these sources: jobs, product, company, survey"
            )
        return url

    def intervals(
        self, from_date: str, to_date: str, days_interval: int
    ) -> Tuple[List[str], List[str]]:
        """
        Breaks dates range into smaller by provided days interval.

        Args:
            from_date (str): Start date for the query in "%Y-%m-%d" format.
            to_date (str): End date for the query, if empty, will be executed as datetime.today().strftime("%Y-%m-%d").
            days_interval (int): Days specified in date range per api call (test showed that 30-40 is optimal for performance).

        Returns:
            List[str], List[str]: Starts and Ends lists that contains information about date ranges for specific period and time interval.

        Raises:
            ValidationError: If the final date of the query is before the start date.
        """

        if to_date == None:
            to_date = datetime.today().strftime("%Y-%m-%d")

        end_date = datetime.strptime(to_date, "%Y-%m-%d").date()
        start_date = datetime.strptime(from_date, "%Y-%m-%d").date()

        from_date_obj = datetime.strptime(from_date, "%Y-%m-%d")

        to_date_obj = datetime.strptime(to_date, "%Y-%m-%d")
        delta = to_date_obj - from_date_obj

        if delta.days < 0:
            raise ValidationError("to_date cannot be earlier than from_date.")

        interval = timedelta(days=days_interval)
        starts = []
        ends = []

        period_start = start_date
        while period_start < end_date:
            period_end = min(period_start + interval, end_date)
            starts.append(period_start.strftime("%Y-%m-%d"))
            ends.append(period_end.strftime("%Y-%m-%d"))
            period_start = period_end
        if len(starts) == 0 and len(ends) == 0:
            starts.append(from_date)
            ends.append(to_date)
        return starts, ends

    def check_connection(
        self,
        source: Literal["jobs", "product", "company", "survey"] = None,
        from_date: str = "2022-03-22",
        to_date: str = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] = "all",
        url: str = None,
    ) -> Tuple[Dict[str, Any], str]:
        """
        Initiate first connection to API to retrieve piece of data with information about type of pagination in API URL.
        This option is added because type of pagination for endpoints is being changed in the future from page number to 'next' id.

        Args:
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None, which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page. 100 entries by default.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional): Region filter for the query. Defaults to "all". [July 2023 status: parameter works only for 'all' on API]
            url (str, optional): Generic part of the URL to Vid Club API. Defaults to None.

        Returns:
            Tuple[Dict[str, Any], str]: Dictionary with first response from API with JSON containing data and used URL string.

        Raises:
            ValidationError: If from_date is earlier than 2022-03-22.
            ValidationError: If to_date is earlier than from_date.
        """

        if from_date < "2022-03-22":
            raise ValidationError("from_date cannot be earlier than 2022-03-22.")

        if to_date < from_date:
            raise ValidationError("to_date cannot be earlier than from_date.")

        if url is None:
            url = self.credentials["url"]

        first_url = self.build_query(
            source=source,
            from_date=from_date,
            to_date=to_date,
            api_url=url,
            items_per_page=items_per_page,
            region=region,
        )
        headers = self.headers
        response = handle_api_response(
            url=first_url, headers=headers, method="GET", verify=False
        )
        response = response.json()

        return (response, first_url)

    def get_response(
        self,
        source: Literal["jobs", "product", "company", "survey"] = None,
        from_date: str = "2022-03-22",
        to_date: str = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] = "all",
    ) -> pd.DataFrame:
        """
        Basing on the pagination type retrieved using check_connection function, gets the response from the API queried and transforms it into DataFrame.

        Args:
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None, which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page. 100 entries by default.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional): Region filter for the query. Defaults to "all". [July 2023 status: parameter works only for 'all' on API]

        Returns:
            pd.DataFrame: Table of the data carried in the response.

        Raises:
            ValidationError: If any source different than the ones in the list are used.
        """
        headers = self.headers
        if source not in ["jobs", "product", "company", "survey"]:
            raise ValidationError(
                "The source has to be: jobs, product, company or survey"
            )
        if to_date == None:
            to_date = datetime.today().strftime("%Y-%m-%d")

        response, first_url = self.check_connection(
            source=source,
            from_date=from_date,
            to_date=to_date,
            items_per_page=items_per_page,
            region=region,
        )

        if isinstance(response, dict):
            keys_list = list(response.keys())
        elif isinstance(response, list):
            keys_list = list(response[0].keys())
        else:
            keys_list = []

        if "next" in keys_list:
            ind = True
        else:
            ind = False

        if "data" in keys_list:
            df = pd.DataFrame(response["data"])
            length = df.shape[0]
            page = 1

            while length == items_per_page:
                if ind == True:
                    next = response["next"]
                    url = f"{first_url}&next={next}"
                else:
                    page += 1
                    url = f"{first_url}&page={page}"
                r = handle_api_response(
                    url=url, headers=headers, method="GET", verify=False
                )
                response = r.json()
                df_page = pd.DataFrame(response["data"])
                if source == "product":
                    df_page = df_page.transpose()
                length = df_page.shape[0]
                df = pd.concat((df, df_page), axis=0)
        else:
            df = pd.DataFrame(response)

        return df

    def total_load(
        self,
        source: Literal["jobs", "product", "company", "survey"] = None,
        from_date: str = "2022-03-22",
        to_date: str = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] = "all",
        days_interval: int = 30,
    ) -> pd.DataFrame:
        """
        Looping get_response and iterating by date ranges defined in intervals. Stores outputs as DataFrames in a list.
        At the end, daframes are concatenated in one and dropped duplicates that would appear when quering.

        Args:
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None, which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page. 100 entries by default.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional): Region filter for the query. Defaults to "all". [July 2023 status: parameter works only for 'all' on API]
            days_interval (int, optional): Days specified in date range per api call (test showed that 30-40 is optimal for performance). Defaults to 30.

        Returns:
            pd.DataFrame: Dataframe of the concatanated data carried in the responses.
        """

        starts, ends = self.intervals(
            from_date=from_date, to_date=to_date, days_interval=days_interval
        )

        dfs_list = []
        if len(starts) > 0 and len(ends) > 0:
            for start, end in zip(starts, ends):
                logger.info(f"ingesting data for dates [{start}]-[{end}]...")
                df = self.get_response(
                    source=source,
                    from_date=start,
                    to_date=end,
                    items_per_page=items_per_page,
                    region=region,
                )
                dfs_list.append(df)
                if len(dfs_list) > 1:
                    df = pd.concat(dfs_list, axis=0, ignore_index=True)
                else:
                    df = pd.DataFrame(dfs_list[0])
        else:
            df = self.get_response(
                source=source,
                from_date=from_date,
                to_date=to_date,
                items_per_page=items_per_page,
                region=region,
            )
        df.drop_duplicates(inplace=True)

        if df.empty:
            logger.error("No data for this date range")

        return df



import base64
import os

# import shutil
import urllib

from github import Github, UnknownObjectException

# import pygit2
from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs

# from typing import Any


# This task is broken on some systems
# class CloneRepo(Task):
#     """
#     Task for cloning a GitHub repository.

#     Args:
#     - repo (str, optional): the name of the repository to clone; must be
#         provided in the form `organization/repo_name` or `user/repo_name`; can also be
#         provided to the `run` method
#     - to (str, optional): the destination folder for the repository; defaults to the repository's name
#     - bare (bool, optional): whether to clone a read-only copy; defaults to `False`

#     Example:
#         ```python
#         from prefect import Flow
#         from viadot.tasks.github import CloneRepo
#         with Flow(name="example") as f:
#             task = CloneRepo()(repo='fishtown-analytics/dbt')
#         out = f.run()
#         ```
#     """

#     def __init__(
#         self,
#         repo: str = None,
#         to: str = None,
#         bare: bool = False,
#         access_token_secret: str = "github_token",
#         **kwargs: Any,
#     ):
#         self.repo = repo
#         self.to = to or repo.split("/")[-1] if repo else to
#         self.bare = bare
#         self.access_token_secret = access_token_secret
#         super().__init__(**kwargs)

#     @defaults_from_attrs("repo", "to", "bare", "access_token_secret")
#     def run(
#         self,
#         repo: str = None,
#         to: str = None,
#         bare: bool = None,
#         access_token_secret: str = None,
#     ):
#         """
#         Clones the repo.

#         Args:
#         - repo (str, optional): the name of the repository to clone; must be
#             provided in the form `organization/repo_name` or `user/repo_name`
#         - to (str, optional): the destination folder for the repository; defaults to the repository's name
#         - bare (bool, optional): whether to clone a read-only copy; defaults to `False`
#         """

#         shutil.rmtree(to, ignore_errors=True)  # Delete folder on run

#         git_token = Secret(access_token_secret).get()
#         repo_url = f"https://{git_token}:x-oauth-basic@github.com/{repo}"

#         if os.path.dirname(to):
#             os.makedirs(os.path.dirname(to), exist_ok=True)
#         pygit2.clone_repository(repo_url, to, bare=bare)

#         self.logger.info(f"Repo {repo} has been successfully cloned.")


class DownloadGitHubFile(Task):
    """
    Task for downloading a file from GitHub.

    Args:
        repo (str, optional): The repository in the format `org/repo`. Defaults to None.
        from_path (str, optional): The path to the file. Defaults to None.
        to_path (str, optional): The destination path. Defaults to None.
        access_token_secret (str, optional): The Prefect secret containing GitHub token. Defaults to "github_token".
        branch (str, optional): The GitHub branch to use. Defaults to "main".
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        repo: str = None,
        from_path: str = None,
        to_path: str = None,
        access_token_secret: str = "github_token",
        branch: str = "main",
        timeout: int = 3600,
        **kwargs,
    ):
        self.repo = repo
        self.from_path = from_path
        self.to_path = to_path
        self.access_token_secret = access_token_secret
        self.branch = branch
        super().__init__(name="download_github_file", timeout=timeout, **kwargs)

    @defaults_from_attrs(
        "repo", "from_path", "to_path", "access_token_secret", "branch"
    )
    def run(
        self,
        repo: str = None,
        from_path: str = None,
        to_path: str = None,
        access_token_secret: str = None,
        branch: str = None,
    ):
        """Task run method.

        Args:
            repo (str, optional): The repository in the format `org/repo`. Defaults to None.
            from_path (str, optional): The path to the file. Defaults to None.
            to_path (str, optional): The destination path. Defaults to None.
            access_token_secret (str, optional): The Prefect secret containing GitHub token. Defaults to "github_token".
            branch (str, optional): The GitHub branch to use. Defaults to "main".
        """
        git_token = Secret(access_token_secret).get()
        file_name = from_path.split("/")[-1]
        to_path = to_path or file_name

        g = Github(git_token)
        repo_obj = g.get_repo(repo)
        try:
            content_encoded = repo_obj.get_contents(
                urllib.parse.quote(from_path), ref=branch
            ).content
        except UnknownObjectException as e:
            full_path = os.path.join(repo, from_path)
            raise ValueError(f"The specified file does not exist under {full_path}.")
        content = base64.b64decode(content_encoded)

        if os.path.dirname(to_path):
            os.makedirs(os.path.dirname(to_path), exist_ok=True)
        with open(to_path, "w") as f:
            f.write(content.decode())



import json
from typing import Any, Dict, List

import pandas as pd
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from ..exceptions import DBDataAccessError
from ..sources import BigQuery
from .azure_key_vault import AzureKeyVaultSecret

logger = logging.get_logger()


class BigQueryToDF(Task):
    """
    The task for querying BigQuery and saving data as the data frame.
    """

    def __init__(
        self,
        query: str = None,
        dataset_name: str = None,
        table_name: str = None,
        start_date: str = None,
        end_date: str = None,
        date_column_name: str = "date",
        credentials_key: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Initialize BigQueryToDF object. For querying on database user have to provide `query` parameter or `dataset` and `table_name`.
        The name of the project is taken from the config/credential json file so there is no need to enter its name.

        If user does not provide `query` there are 3 cases:
            If `start_date` and `end_date` are not None - all data from the start date to the end date will be retrieved.
            If `start_date` and `end_date` are left as default (None) - the data is pulled till "yesterday" (current date -1)
            If the column that looks like a date does not exist in the table, get all the data from the table.

        Args:
            query(str, optional): SQL query to querying data in BigQuery. Format of basic query: (SELECT * FROM `{project}.{dataset_name}.{table_name}`)
                Defaults to None.
            dataset_name (str, optional): Dataset name. Defaults to None.
            table_name (str, optional): Table name. Defaults to None.
            date_column_name (str, optional): The query is based on a date, the user can provide the name
                of the date columnn if it is different than "date". If the user-specified column does not exist,
                all data will be retrieved from the table. Defaults to "date".
            start_date (str, optional): A query parameter to pass start date e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A query parameter to pass end date e.g. "2022-01-01". Defaults to None.
            credentials_key (str, optional): Credential key to dictionary where details are stored - Azure Key Vault or local config.
                credentials can be generated as key for User Principal inside a BigQuery project. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.query = query
        self.dataset_name = dataset_name
        self.table_name = table_name
        self.start_date = start_date
        self.end_date = end_date
        self.date_column_name = date_column_name
        self.credentials_key = credentials_key
        self.vault_name = vault_name

        super().__init__(
            name="bigquery_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download BigQuery data to a DF"""
        super().__call__(self)

    @defaults_from_attrs(
        "query",
        "dataset_name",
        "table_name",
        "date_column_name",
        "start_date",
        "end_date",
        "credentials_key",
        "vault_name",
    )
    def run(
        self,
        query: str = None,
        dataset_name: str = None,
        table_name: str = None,
        date_column_name: str = "date",
        start_date: str = None,
        end_date: str = None,
        credentials_key: str = None,
        vault_name: str = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        credentials = None
        if credentials_key:
            credentials_str = AzureKeyVaultSecret(
                secret=credentials_key, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)

        bigquery = BigQuery(credentials_key=credentials_key, credentials=credentials)
        project = bigquery.get_project_id()
        if query is None:
            try:
                table_columns = bigquery.list_columns(
                    dataset_name=dataset_name, table_name=table_name
                )
                if date_column_name not in table_columns:
                    logger.warning(
                        f"'{date_column_name}' column is not recognized. Downloading all the data from '{table_name}'."
                    )
                    query = f"SELECT * FROM `{project}.{dataset_name}.{table_name}`"
                    df = bigquery.query_to_df(query)
                else:
                    if start_date is not None and end_date is not None:
                        query = f"""SELECT * FROM `{project}.{dataset_name}.{table_name}` 
                        where {date_column_name} between PARSE_DATE("%Y-%m-%d", "{start_date}") and PARSE_DATE("%Y-%m-%d", "{end_date}") 
                        order by {date_column_name} desc"""
                    else:
                        query = f"""SELECT * FROM `{project}.{dataset_name}.{table_name}` 
                        where {date_column_name} < CURRENT_DATE() 
                        order by {date_column_name} desc"""

                    df = bigquery.query_to_df(query)
                    logger.info(
                        f"Downloaded the data from the '{table_name}' table into the data frame."
                    )

            except DBDataAccessError:
                logger.warning(
                    f"Table name '{table_name}' or dataset '{dataset_name}' not recognized. Returning empty data frame."
                )
                df = pd.DataFrame()
            return df
        else:
            try:
                df = bigquery.query_to_df(query)
            except Exception as e:
                logger.warning(
                    f"The query is invalid. Please enter a valid query. Returning an empty dataframe."
                )
                df = pd.DataFrame()
            return df



import pytest

from viadot.config import local_config
from viadot.exceptions import DataRangeError
from viadot.sources import Epicor


@pytest.fixture(scope="session")
def epicor():
    epicor = Epicor(
        base_url=local_config.get("EPICOR").get("test_url"),
        config_key="EPICOR",
        filters_xml="""
    <OrderQuery>
        <QueryFields>
            <CompanyNumber>001</CompanyNumber>
            <BegInvoiceDate>2022-05-16</BegInvoiceDate>
            <EndInvoiceDate>2022-05-16</EndInvoiceDate>
            <RecordCount>3</RecordCount>
        </QueryFields>
    </OrderQuery>""",
    )
    yield epicor


@pytest.fixture(scope="session")
def epicor_error():
    epicor_error = Epicor(
        base_url=local_config.get("EPICOR").get("test_url"),
        config_key="EPICOR",
        filters_xml="""
    <OrderQuery>
        <QueryFields>
            <CompanyNumber>001</CompanyNumber>
            <BegInvoiceDate></BegInvoiceDate>
            <EndInvoiceDate>2022-05-16</EndInvoiceDate>
            <RecordCount>3</RecordCount>
        </QueryFields>
    </OrderQuery>""",
    )
    yield epicor_error


def test_connection(epicor):
    assert epicor.get_xml_response().ok


def test_validate_filter(epicor_error):
    with pytest.raises(DataRangeError):
        epicor_error.validate_filter()



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import SAPBWToADLS

DATA = {
    "[0CALMONTH].[LEVEL01].[DESCRIPTION]": ["January 2023"],
    "date": ["2023-06-19 11:12:43+00:00"],
}

ADLS_FILE_NAME = "test_sap_bw_to_adls.parquet"
ADLS_DIR_PATH = "raw/tests/"


@mock.patch(
    "viadot.tasks.SAPBWToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_sap_bw_to_adls_flow_run(mocked_class):
    flow = SAPBWToADLS(
        "test_sap_bw_to_adls_flow_run",
        sapbw_credentials_key="SAP",
        env="BW",
        mdx_query="""
            SELECT
                    {
                }
                    ON COLUMNS,
            NON EMPTY
                    { 
                        { [0CALMONTH].[202301] } 
            } 
            DIMENSION PROPERTIES
            DESCRIPTION,
            MEMBER_NAME
            ON ROWS

            FROM ZCSALORD1/ZBW4_ZCSALORD1_006_BOA
        
                    """,
        mapping_dict={"[0CALMONTH].[LEVEL01].[DESCRIPTION]": "Calendar Year/Month"},
        overwrite_adls=True,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_sap_bw_to_adls_flow_run.parquet")
    os.remove("test_sap_bw_to_adls_flow_run.json")



import uuid

import pytest
from azure.core.exceptions import ResourceNotFoundError

from viadot.tasks import (
    AzureKeyVaultSecret,
    CreateAzureKeyVaultSecret,
    DeleteAzureKeyVaultSecret,
)

uuid4 = uuid.uuid4()
SECRET_NAME = f"test-secret-{uuid4}"
SECRET_VALUE = "test_value"


def test_create_azure_key_vault_secret():
    create_azure_key_vault_task = CreateAzureKeyVaultSecret()
    was_successful = create_azure_key_vault_task.run(
        secret=SECRET_NAME, value=SECRET_VALUE, lifetime=0.0005
    )
    assert was_successful


def test_read_azure_key_vault_secret():
    azure_key_vault_task = AzureKeyVaultSecret()
    value = azure_key_vault_task.run(secret=SECRET_NAME)
    assert value == SECRET_VALUE


def test_delete_azure_key_vault_secret():
    delete_azure_key_vault_task = DeleteAzureKeyVaultSecret()
    was_successful = delete_azure_key_vault_task.run(secret=SECRET_NAME)

    assert was_successful

    with pytest.raises(ResourceNotFoundError):
        azure_key_vault_task = AzureKeyVaultSecret()
        azure_key_vault_task.run(secret=SECRET_NAME)



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import ADLSGen1ToAzureSQLNew

d = {"country": [1, 2], "sales": [3, 4]}
df = pd.DataFrame(data=d)
SCHEMA = "sandbox"
TABLE = "test_bcp"


def test_adls_gen1_to_azure_sql_new_init_args():
    flow = ADLSGen1ToAzureSQLNew(
        name="test_adls_gen1_gen2_flow",
        gen1_path="test_file_1.csv",
        gen2_path="test_file_2.csv",
        schema=SCHEMA,
        table=TABLE,
        dtypes={"country": "INT", "sales": "INT"},
        if_exists="replace",
    )

    assert flow


def test_adls_gen1_to_azure_sql_new_mock():
    with mock.patch.object(
        ADLSGen1ToAzureSQLNew, "run", return_value=True
    ) as mock_method:
        instance = ADLSGen1ToAzureSQLNew(
            name="test_adls_gen1_gen2_flow",
            gen1_path="folder1/example_file.csv",
            gen2_path="folder2/example_file.csv",
            schema="sandbox",
            table="test_bcp",
            dtypes={"country": "VARCHAR(25)", "sales": "INT"},
            if_exists="replace",
        )
        instance.run()
        mock_method.assert_called_with()


@mock.patch(
    "viadot.tasks.AzureDataLakeToDF.run",
    return_value=pd.DataFrame(data={"country": [1, 2], "sales": [3, 4]}),
)
@pytest.mark.run
def test_adls_gen1_to_azure_sql_new_flow_run_mock(mocked_class):
    flow = ADLSGen1ToAzureSQLNew(
        name="test_adls_g1g2",
        gen1_path="example_path",
        gen2_path="raw/test/test.csv",
        dtypes={"country": "VARCHAR(25)", "sales": "INT"},
        if_exists="replace",
        table="test",
        schema="sandbox",
    )

    result = flow.run()

    assert result.is_successful()
    os.remove("test_adls_g1g2.csv")



from typing import Any, Dict, List

from prefect import Flow, Task, apply_map
from prefect.tasks.prefect import StartFlowRun


class Pipeline(Flow):
    def __init__(
        self,
        name: str,
        project_name: str,
        extract_flows_names: List[str],
        transform_flow_name: str,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        self.extract_flows_names = extract_flows_names
        self.transform_flow_name = transform_flow_name
        self.project_name = project_name
        self.timeout = timeout
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    def gen_start_flow_run_task(self, flow_name: str, flow: Flow = None) -> Task:
        start_flow_run_task = StartFlowRun(wait=True, timeout=self.timeout)
        t = start_flow_run_task.bind(
            flow_name=flow_name, project_name=self.project_name, flow=flow
        )
        return t

    def gen_flow(self):
        extract_flow_runs = apply_map(
            self.gen_start_flow_run_task, self.extract_flows_names, flow=self
        )
        start_flow_run_task_2 = StartFlowRun(wait=True, timeout=self.timeout)
        transform_flow_run = start_flow_run_task_2.bind(
            flow_name=self.transform_flow_name,
            project_name=self.project_name,
            flow=self,
        )
        transform_flow_run.set_upstream(extract_flow_runs, flow=self)



import datetime
from datetime import date

import pandas as pd
import pytest

from viadot.tasks import GetFlowNewDateRange
from viadot.tasks.prefect_date_range import (
    calculate_difference,
    check_if_scheduled_run,
    get_formatted_date,
    get_time_from_last_successful_run,
)

PREFECT_TASK = GetFlowNewDateRange()
DATE_FROM_PREFECT = "2022-01-01T01:30:00+00:00"
DATE_FROM_PREFECT2 = "2022-01-04T02:20:00+00:00"
PREFECT_JSON = {
    "data": {
        "flow": [
            {
                "flow_runs": [
                    {
                        "id": "88a505ec-f7c9-4242-baa4-81c9db1aaafe",
                        "start_time": "2022-02-21T01:04:44.146191+00:00",
                        "state": "Failed",
                        "scheduled_start_time": "2022-02-21T01:00:00+00:00",
                    },
                    {
                        "id": "9d310a47-2fb0-4435-971c-4fcaf3efca06",
                        "start_time": "2022-02-20T01:05:36.142547+00:00",
                        "state": "Success",
                        "scheduled_start_time": "2022-02-20T01:00:00+00:00",
                    },
                    {
                        "id": "90c32b13-4a0a-467b-a07f-c568a34f1bc2",
                        "start_time": "2022-02-19T01:05:30.572431+00:00",
                        "state": "Failed",
                        "scheduled_start_time": "2022-02-19T01:00:00+00:00",
                    },
                ]
            }
        ]
    }
}


def test_get_formatted_dated():
    new_date = get_formatted_date(time_unclean=DATE_FROM_PREFECT, return_value="date")
    assert new_date == datetime.date(2022, 1, 1)
    assert isinstance(new_date, date)

    new_time = get_formatted_date(time_unclean=DATE_FROM_PREFECT, return_value="time")
    assert new_time == datetime.time(1, 30)
    assert isinstance(new_time, datetime.time)


def test_calculate_difference_date():
    diff_days = calculate_difference(
        date_to_compare=DATE_FROM_PREFECT2,
        base_date=DATE_FROM_PREFECT,
        diff_type="date",
    )
    assert diff_days == 3


def test_calculate_difference_time():
    diff_hours = calculate_difference(
        date_to_compare=DATE_FROM_PREFECT2,
        base_date=DATE_FROM_PREFECT,
        diff_type="time",
    )
    assert diff_hours == 0

    diff_hours = calculate_difference(
        date_to_compare="2022-01-04T02:50:00+00:00",
        base_date=DATE_FROM_PREFECT,
        diff_type="time",
    )
    assert diff_hours == 1.20


def test_get_time_from_last_successful_run():
    flow_runs = PREFECT_JSON["data"]["flow"]
    start_time_success = get_time_from_last_successful_run(flow_runs_details=flow_runs)
    assert start_time_success == "2022-02-20T01:05:36.142547+00:00"


def test_check_if_scheduled_run():
    is_scheduled = check_if_scheduled_run(
        time_run="2022-02-21T01:40:00+00:00", time_schedule="2022-02-15T01:00:00+00:00"
    )
    assert is_scheduled is True

    is_scheduled = check_if_scheduled_run(
        time_run="2022-02-21T02:20:00+00:00", time_schedule="2022-02-15T01:00:00+00:00"
    )
    assert is_scheduled is False


def test_change_date_range():
    date_range = "last_5_days"
    difference = 10
    date_range_new = PREFECT_TASK.change_date_range(
        date_range=date_range, difference=difference
    )
    assert date_range_new == "last_15_days"



from viadot.sources.azure_blob_storage import AzureBlobStorage


def test_if_exists():
    azstorage = AzureBlobStorage()
    assert azstorage.exists("tests/test.csv")


def test_to_storage(TEST_CSV_FILE_PATH, TEST_CSV_FILE_BLOB_PATH):
    azstorage = AzureBlobStorage()
    azstorage.to_storage(
        from_path=TEST_CSV_FILE_PATH, to_path=TEST_CSV_FILE_BLOB_PATH, overwrite=True
    )


#    file = AzureBlobStorage(path="/tests/test.png")
#    assert file.exists
